/*
   File: affix_rules.c
   Checks the affix rules

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: affix_rules.c,v 1.22 2011/09/02 10:11:23 marcs Exp $"
*/

/* standard includes */
#include <stdio.h>
#include <string.h>

/* support lib includes */
#include <dcg.h>
#include <dcg_alloc.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "cdl_ds.h"
#include "options.h"
#include "globals.h"
#include "modules.h"
#include "affix_tree.h"
#include "affix_rules.h"
#include "warshall.h"
#include "contsens.h"

/*
   All affix rules will be stored in one big list which can be indexed by affix_rule_nr.
   We need this indexing anyway as the warshall algorithm will be needed for chain rule
   detection.
*/
static int unique_affix_rule_nr;

/*
   Phase 0:
   Initialize list of all affix rules
*/
void init_affix_rules ()
{ unique_affix_rule_nr = 0;
}

string new_anonymous_affix_name ()
{ string aname = dcg_new_fmtd_string ("_ANON__%d", unique_affix_rule_nr);
  unique_affix_rule_nr++;
  return (aname);
}

/*
   Affix rule checking phase 1
   Register all affix rule definitions into a local affix tree
   and all_affix_rules list and put the self ref into the affix rule.
   Then identify all exported definitions: we do not allow reexports
*/
static void register_affix_rules_for_module (module mod)
{ affix_rule_list arules = mod -> arules;
  int ix;

  /* Enter all affix rules into the local affix tree */
  for (ix = 0; ix < arules -> size; ix++)
    { affix_rule arule = arules -> array[ix];
      arule -> anr = all_affix_rules -> size;
      arule -> mnr = mod -> mnr;
      app_affix_rule_list (all_affix_rules, arule);
      if (!enter_affix_tree (&mod -> adefs, arule))
	contsens_error (mod, arule -> line, arule -> col,
			"multiple definition of affix rule %s", arule -> aname);
    };
}

static void export_affix_rules_for_module (module mod)
{ /* Identify exported affix rules */
  export_list defines = mod -> defines;
  int ix;
  if (partial_compilation && (mod == root_module))
    defines = root_module -> Root_module.exports;
  for (ix = 0; ix < defines -> size; ix++)
    { export exp = defines -> array[ix];
      affix_rule adef;
      string aname;
      if (exp -> tag != TAGExport_affix) continue;
      aname = exp -> Export_affix.aname;
      adef = lookup_affix_tree (mod -> adefs, aname);
      if (adef == affix_rule_nil)
	contsens_error (mod, exp -> line, exp -> col, "can't identify affix rule %s", aname);
      else exp -> Export_affix.adef = adef;
    };
}

static void register_and_export_affix_rules ()
{ int ix;
  dcg_hint ("      registering and exporting affix rules");
  for (ix = 0; ix < all_modules -> size; ix++)
    { module mod = all_modules -> array[ix];
      register_affix_rules_for_module (mod);
      export_affix_rules_for_module (mod);
    };
}

/*
   Affix rule checking phase 2
   Import exported affix rule definitions into the modules using the uses list.
   Following the import, we identify all affixes in the right hand sides of affix rules
*/
static void import_affix_rules_from_module (module imp, module mod)
{ export_list defines = imp -> defines;
  int ix;
  for (ix = 0; ix < defines -> size; ix++)
    { export exp = defines -> array[ix];
      affix_rule import;
      if (exp -> tag != TAGExport_affix) continue;

      /* Imports are added to the module's affix tree, but not to the list of all affix rules */
      import = new_Affix_import (exp -> line, exp -> col,
				 attach_string (exp -> Export_affix.aname));
      import -> Affix_import.from = imp;
      import -> Affix_import.impdef = exp -> Export_affix.adef;

      if (!enter_affix_tree (&mod -> adefs, import))
	contsens_error (mod, 0, 0, "import of affix rule %s from module %s conflicts",
			import -> aname, imp -> mname);
      else app_affix_rule_list (mod -> arules, import);

      /* Mark the import affix rule */
      import -> anr = -1;
    };
}

static void import_affix_rules_in_module (module mod)
{ string_list uses = mod -> uses;
  int ix;
  for (ix = 0; ix < uses -> size; ix++)
    { string uname = uses -> array[ix];
      module imp = identify_module (uname);
      import_affix_rules_from_module (imp, mod);
    };
}

/*
   Phase 2b: Identify affixes in right hand sides
   For affix names that are imported, the corresponding affix rule in some
   other module is identified.
*/
static affix_rule try_identify_affix_rule (module mod, int line, int col, string aname)
{ affix_rule adef = lookup_affix_tree (mod -> adefs, aname);
  if (adef == affix_rule_nil)
    contsens_error (mod, line, col, "affix rule %s can't be identified", aname);
  else if (adef -> tag == TAGAffix_import)
    return (adef -> Affix_import.impdef);
  return (adef);
}

static void identify_affix_synonym_definition (affix_rule rule, module mod)
{ affix_rule syndef = try_identify_affix_rule (mod, rule -> line, rule -> col,
					       rule -> Affix_synonym.syn);
  rule -> Affix_synonym.syndef = syndef;
}

static void identify_affix_element (affix_element elem, module mod)
{ affix_rule vdef;
  if (elem -> tag != TAGAffix_var) return;	/* markers not identified */
  vdef = try_identify_affix_rule (mod, elem -> line, elem -> col, elem -> Affix_var.vname);
  elem -> Affix_var.vdef = vdef;
}

static void identify_affixes_in_affix_elements (affix_element_list elems, module mod)
{ int ix;
  for (ix = 0; ix < elems -> size; ix++)
    identify_affix_element (elems -> array[ix], mod);
}

static void identify_affixes_in_affix_alts (affix_alternative_list alts, module mod)
{ int ix;
  for (ix = 0; ix < alts -> size; ix++)
    identify_affixes_in_affix_elements (alts -> array[ix] -> elems, mod);
}

static void identify_affixes_in_rule (affix_rule rule, module mod)
{ switch (rule -> tag)
    { case TAGAffix_synonym:
	identify_affix_synonym_definition (rule, mod);
	break;
      case TAGAffix_alts:
	identify_affixes_in_affix_alts (rule -> Affix_alts.alts, mod);
	break;
      case TAGAffix_abstr:
	identify_affixes_in_affix_alts (rule -> Affix_abstr.alts, mod);
	break;
      case TAGAffix_import:	/* identified earlier */
      case TAGAffix_prim:	/* done */
      default: break;
    };
}

static void identify_affix_rules_in_module (module mod)
{ affix_rule_list arules = mod -> arules;
  int ix;
  for (ix = 0; ix < arules -> size; ix++)
    identify_affixes_in_rule (arules -> array[ix], mod);
}
 
static void import_and_identify_affix_rules_in_module (module mod)
{ import_affix_rules_in_module (mod);
  identify_affix_rules_in_module (mod);
}

static void import_and_identify_affix_rules ()
{ int ix;
  dcg_hint ("      importing and identifying affix rules");
  for (ix = 0; ix < all_modules -> size; ix++)
    import_and_identify_affix_rules_in_module (all_modules -> array[ix]);
}

/*
   Phase 3: Detect all chain productions in the affix rule system
   Iterate over all affix rules to determine which rules contain
   a chain production i.e. a rule of the form N :: N2; or N = N2; 
   Note that this includes synonym definitions.

   Initially we determine the one step chain production as follows:
   rel [i,j] = 1 <=> affix rule i has a chain production to affix rule j.
   Using warshall, we then determine the transitive closure of this
   relation. Those rules who have a 1 on their diagonal entry in the
   transitive closure, contain a chain production in one or more steps
   to itself, which is erroneous.
*/
static void mark_chain_productions_in_alts (affix_alternative_list alts, int aff_nr,
					    int asize, char *rel)
{ int ix;
  for (ix = 0; ix < alts -> size; ix++)
    { affix_element_list elts = alts -> array[ix] -> elems;
      affix_element elt;
      if (elts -> size != 1)	/* at least >1 marker or affix variable */
	continue;
      elt = elts -> array[0];
      if (elt -> tag == TAGAffix_var)
	{ /* We really have a chain production */
	  int cnr = elt -> Affix_var.vdef -> anr;
	  if ((0 <= cnr) && (cnr < asize))
	    rel[aff_nr * asize + cnr] = 1;
	  else dcg_internal_error ("mark_chain_productions_in_alts");
	};
    };
}

static void mark_chain_productions_in_affix_rule (affix_rule arule, int asize, char *rel)
{ int aff_nr = arule -> anr;
  switch (arule -> tag)
    { case TAGAffix_synonym:
	{ /* Synonyms are always chain productions */
	  affix_rule syn = arule -> Affix_synonym.syndef;
	  int syn_nr = syn -> anr;
	  rel[aff_nr * asize + syn_nr] = 1;
	}; break;
      case TAGAffix_alts:
	{ affix_alternative_list alts = arule -> Affix_alts.alts;
	  mark_chain_productions_in_alts (alts, aff_nr, asize, rel);
        }; break;
      case TAGAffix_abstr:
	{ affix_alternative_list alts = arule -> Affix_abstr.alts;
	  mark_chain_productions_in_alts (alts, aff_nr, asize, rel);
	};
      case TAGAffix_import:
      case TAGAffix_prim: break;
      default: dcg_bad_tag (arule -> tag, "mark_chain_productions_in_affix_rule");
    };
}

static void report_chain_affix_rules_in_module (module mod, int asize, char *trans)
{ affix_rule_list arules = mod -> arules;
  int ix;
  for (ix = 0; ix < arules -> size; ix++)
    { affix_rule arule = arules -> array[ix];
      int anr = arule -> anr;
      if (anr < 0) continue;		/* Imported definition */
      if (trans[anr * asize + anr])
	contsens_error (mod, arule -> line, arule -> col,
			"affix rule %s defines a chain production", arule -> aname);
    };
}

static void detect_chain_productions ()
{ int ix;
  int asize = all_affix_rules -> size;
  char *rel = dcg_calloc (asize * asize, sizeof (char));
  char *trans;

  /* Mark the one step chain productions */
  for (ix = 0; ix < asize * asize; ix++) rel[ix] = 0;
  dcg_hint ("      detecting chain productions in affix rules");
  for (ix = 0; ix < asize; ix++)
    { affix_rule arule = all_affix_rules -> array[ix];
      mark_chain_productions_in_affix_rule (arule, asize, rel);
    };

  /* Report the multi step self chain productions */
  trans = warshall (asize, rel); 
  for (ix = 0; ix < all_modules -> size; ix++)
    report_chain_affix_rules_in_module (all_modules -> array[ix], asize, trans);

  /* Free the two arrays: we do not need them */
  dcg_detach ((void **) &rel);
  dcg_detach ((void **) &trans);
}

/*
   Phase 4: Resolution of all synonym and import types.
   Following this phase all syndefs or impdefs point to their final equivalent
*/
static int is_a_synonym_affix_rule (affix_rule arule, affix_rule *sdef)
{ if (arule == affix_rule_nil) return (0);
  switch (arule -> tag)
    { case TAGAffix_synonym: *sdef = arule -> Affix_synonym.syndef; break;
      case TAGAffix_import:  *sdef = arule -> Affix_import.impdef; break;
      default: return (0);
    };
  return (1);
}

static void set_affix_synonym_def (affix_rule arule, affix_rule sdef)
{ switch (arule -> tag)
    { case TAGAffix_synonym: arule -> Affix_synonym.syndef = sdef; break;
      case TAGAffix_import:  arule -> Affix_import.impdef =  sdef; break;
      default: dcg_bad_tag (arule -> tag, "set_affix_synonym_def");
    };
}
  
static affix_rule try_resolve_synonym_rule (affix_rule arule)
{ affix_rule syndef, syndef2;

  /* If we are not some synonym, we're done */
  if (!is_a_synonym_affix_rule (arule, &syndef))
    return (arule);

  /* Determine if we are a synonym of something else */
  /* Note: if we have a chain of synonyms, it is traversed only once */
  syndef2 = try_resolve_synonym_rule (syndef);
  set_affix_synonym_def (arule, syndef2);
  return (syndef2);
}

static void resolve_synonym_rules_in_module (module mod)
{ affix_rule_list arules = mod -> arules;
  int ix;
  for (ix = 0; ix < arules -> size; ix++)
    (void) try_resolve_synonym_rule (arules -> array[ix]);
}

static void resolve_synonym_rules ()
{ int ix;
  dcg_hint ("      resolving synonym and imported affix rules");
  for (ix = 0; ix < all_modules -> size; ix++)
    resolve_synonym_rules_in_module (all_modules -> array[ix]);
}

/*
   Phase 5: check some wellformedness
   Two alternatives of the same affix rule may not be identical
*/
static int same_element (affix_element elem1, affix_element elem2)
{ if (elem1 -> tag != elem2 -> tag) return (0);
  switch (elem1 -> tag)
    { case TAGAffix_var:
	return (equivalent_affix_rule (elem1 -> Affix_var.vdef, elem2 -> Affix_var.vdef));
      case TAGAffix_term:
	return (streq (elem1 -> Affix_term.tname, elem2 -> Affix_term.tname));
      default: return (1);
    };
}

static int same_alternatives (affix_alternative alt1, affix_alternative alt2)
{ affix_element_list elems1 = alt1 -> elems;
  affix_element_list elems2 = alt2 -> elems;
  int ix;
  if (elems1 -> size != elems2 -> size) return (0);
  for (ix = 0; ix < elems1 -> size; ix++)
    if (!same_element (elems1 -> array[ix], elems2 -> array[ix])) return (0);
  return (1);
}

static void check_wellformed_alternatives (affix_rule arule, affix_alternative_list alts,
					   module mod)
{ int ix, iy;
  for (ix = 0; ix < alts -> size; ix++)
    for (iy = ix + 1; iy < alts -> size; iy++)
      if (same_alternatives (alts -> array[ix], alts -> array[iy]))
	contsens_error (mod, arule -> line, arule -> col,
			"alternative %d and alternative %d of affix rule are equal",
			ix + 1, iy + 1, arule -> aname);
}

static void check_wellformedness_in_module (module mod)
{ affix_rule_list arules = mod -> arules;
  int ix;
  for (ix = 0; ix < arules -> size; ix++)
    { affix_rule arule = arules -> array[ix];
      if (arule -> tag == TAGAffix_alts)
	check_wellformed_alternatives (arule, arule -> Affix_alts.alts, mod);
    };
}

static void check_wellformedness ()
{ int ix;
  dcg_hint ("      checking wellformedness conditions");
  for (ix = 0; ix < all_modules -> size; ix++)
    check_wellformedness_in_module (all_modules -> array[ix]);
}

/*
   Following phase 5, the type check is complete:
   The following routines may then be used for type access and comparison
*/
affix_rule get_prime_affix_rule (affix_rule arule)
{ affix_rule sdef;
  if (arule == affix_rule_nil) return (arule);
  if (arule == AFFIX_ERROR) return (AFFIX_ERROR);
  if (is_a_synonym_affix_rule (arule, &sdef)) return (sdef);
  return (arule);
}

string get_affix_rule_name (affix_rule arule)
{ affix_rule sdef = get_prime_affix_rule (arule);
  return (sdef -> aname);
}

affix_alternative_list pick_rhs_affix_alts (affix_rule arule)
{ affix_rule sdef = get_prime_affix_rule (arule);
  if ((sdef == affix_rule_nil) || (sdef == AFFIX_ERROR))
    return (affix_alternative_list_nil);
  switch (sdef -> tag)
    { case TAGAffix_alts:  return (sdef -> Affix_alts.alts);
      case TAGAffix_abstr: return (sdef -> Affix_abstr.alts);
      default: return (affix_alternative_list_nil);
    };
}

/*
   Check if a certain rule is an abstract synonym i.e.
   a rule of the form N :: P
*/
int is_abstract_synonym (affix_rule rule, affix_rule *orig_arule, affix_rule *rhs)
{ affix_rule lhs = get_prime_affix_rule (rule);
  affix_rule concr;

  if (lhs -> tag != TAGAffix_abstr) return (0);
  concr = lhs -> Affix_abstr.alts -> array[0] -> elems -> array[0] -> Affix_var.vdef;
  *orig_arule = lhs;
  *rhs = get_prime_affix_rule (concr);
  return (1);
}

int equivalent_affix_rule (affix_rule arule1, affix_rule arule2)
{ affix_rule sdef1 = get_prime_affix_rule (arule1);
  affix_rule sdef2 = get_prime_affix_rule (arule2);
  if (sdef1 == affix_rule_nil) return (0);
  if (sdef2 == affix_rule_nil) return (0);
  if (sdef1 == AFFIX_ERROR) return (0);
  if (sdef2 == AFFIX_ERROR) return (0);
  return (sdef1 == sdef2);
}

int equivalent_signatures (affix_rule_list arules1, affix_rule_list arules2)
{ int ix;
  if (arules1 -> size != arules2 -> size) return (0);
  for (ix = 0; ix < arules1 -> size; ix++)
    if (!equivalent_affix_rule (arules1 -> array[ix], arules2 -> array[ix]))
      return (0);
  return (1);
}

/*
   Phase 6:
   Determine the affix rule representations
   Determine the leftcorner relation between the affix nonterminals
*/
static int enumerable_elements (affix_element_list elems)
{ int ix;
  for (ix = 0; ix < elems -> size; ix++)
    if (elems -> array[ix] -> tag != TAGAffix_term)
      return (0);
  return (1);
}

static int enumerable_alts (affix_alternative_list alts)
{ int ix;
  for (ix = 0; ix < alts -> size; ix++)
    if (!enumerable_elements (alts -> array[ix] -> elems))
      return (0);
  return (1);
}

static void determine_enumerable_types_in_module (module mod)
{ affix_rule_list arules = mod -> arules;
  int ix;
  for (ix = 0; ix < arules -> size; ix++)
    { affix_rule arule = arules -> array[ix];
      if (arule -> tag != TAGAffix_alts) continue;
      arule -> Affix_alts.enumerable = enumerable_alts (arule -> Affix_alts.alts);
    };
}

static void determine_representations_in_module (module mod)
{ affix_rule_list arules = mod -> arules;
  int ix;
  for (ix = 0; ix < arules -> size; ix++)
    { affix_rule arule = arules -> array[ix];
      affix_rule repr = arule;
      affix_rule orig_arule;
      int enumerable = 0;
      int needs_gc = 0;
      int found = 0;
      while (!found)
	{ if (is_abstract_synonym (repr, &orig_arule, &repr)) ;
	  else if (is_a_synonym_affix_rule (repr, &repr)) ;
	  else found = 1;
	};

      switch (arule -> tag)
	{ case TAGAffix_abstr: arule -> Affix_abstr.fin_repr = repr; break;
	  case TAGAffix_import: arule -> Affix_import.fin_repr = repr; break;
	  case TAGAffix_synonym: arule -> Affix_synonym.fin_repr = repr;
	  default: break;
	};

      switch (repr -> tag)
	{ case TAGAffix_alts:
	    enumerable = repr -> Affix_alts.enumerable;
	    needs_gc = !enumerable;
	    break;
	  case TAGAffix_prim:
	    if (!streq (repr -> aname, "INT") && !streq (repr -> aname, "REAL"))
	      needs_gc = 1;
	    break;
	  default: dcg_bad_tag (repr -> tag, "determine_representations_in_module");
	};
      arule -> needs_gc = needs_gc;

      if (dump_affix_rules && (arule -> tag != TAGAffix_import))
	{ if (needs_gc)
	    dcg_hint ("         Affix rule %s needs garbage collection", arule -> aname);
	  else if (enumerable)
	    dcg_hint ("         Affix rule %s is enumerable", arule -> aname);
	  else dcg_hint ("         Affix rule %s does not need garbage collection", arule -> aname);
	};
    };
}

affix_rule get_repr_affix_rule (affix_rule arule)
{ switch (arule -> tag)
    { case TAGAffix_abstr:   return (arule -> Affix_abstr.fin_repr);
      case TAGAffix_import:  return (arule -> Affix_import.fin_repr);
      case TAGAffix_synonym: return (arule -> Affix_synonym.fin_repr);
      case TAGAffix_alts:
      case TAGAffix_prim:    return (arule);
      default: dcg_bad_tag (arule -> tag, "get_repr_affix_rule");
    };
  return (affix_rule_nil);
}

static void determine_representations ()
{ int ix;
  dcg_hint ("      determining representations");
  for (ix = 0; ix < all_modules -> size; ix++)
    determine_enumerable_types_in_module (all_modules -> array[ix]);
  for (ix = 0; ix < all_modules -> size; ix++)
    determine_representations_in_module (all_modules -> array[ix]);
}

static char *determine_direct_leftcorner_relation ()
{ int asize = all_affix_rules -> size;
  char *rel = (char *) dcg_calloc (asize * asize, sizeof (char));
  int ix, iy;
  for (ix = 0; ix < asize * asize; ix++) rel[ix] = 0;
  for (ix = 0; ix < asize; ix++)
    { affix_rule arule = all_affix_rules -> array[ix];
      affix_alternative_list alts;
      switch (arule -> tag)
	{ case TAGAffix_alts:  alts = arule -> Affix_alts.alts; break;
	  case TAGAffix_abstr: alts = arule -> Affix_abstr.alts; break;
	  default: continue;
	};
      for (iy = 0; iy < alts -> size; iy++)
	{ affix_element left_elem = alts -> array[iy] -> elems -> array[0];
	  if (left_elem -> tag == TAGAffix_var)
	    { /* Pick number through synonyms, do not mark primitive types as LC's */
	      affix_rule lc_rule = get_prime_affix_rule (left_elem -> Affix_var.vdef);
	      int lc_rule_nr = lc_rule -> anr;
	      if ((lc_rule -> tag == TAGAffix_abstr) || (lc_rule -> tag == TAGAffix_alts))
	        rel[ix * asize + lc_rule_nr] |= 1;
	    };
	};
    };
  return (rel);
}

static void determine_leftcorner_relation ()
{ int asize = all_affix_rules -> size;
  char *direct_lc, *refl_trans_lc;
  int ix, iy;
  dcg_hint ("      determining affix rule leftcorner relation");
  direct_lc = determine_direct_leftcorner_relation (&direct_lc);
  refl_trans_lc = warshall (asize, direct_lc);
  for (ix = 0; ix < asize; ix++)
    refl_trans_lc[ix * asize + ix] |= 1;
  
  /* Annotate all rules with the all ids of the affix rules that are a left corner */
  for (ix = 0; ix < asize; ix++)
    { affix_rule arule = all_affix_rules -> array[ix];
      if ((arule -> tag != TAGAffix_abstr) && (arule -> tag != TAGAffix_alts))
	arule -> lc_nrs = int_list_nil;
      else
	{ arule -> lc_nrs = new_int_list ();
	  for (iy = 0; iy < asize; iy++)
	    if (refl_trans_lc[ix * asize + iy])
	      app_int_list (arule -> lc_nrs, iy);
	};
    };
}

/*
   Lookup and static storage of the primitive types INT, REAL, STRING
*/
static affix_rule int_arule;
static affix_rule real_arule;
static affix_rule text_arule;
static void create_primitive_type_refs ()
{ module predef = identify_module (PREDEF_NAME);
  int_arule = lookup_affix_tree (predef -> adefs, "INT");
  if (int_arule == affix_rule_nil)
    dcg_panic ("could not locate primitive type INT");
  real_arule = lookup_affix_tree (predef -> adefs, "REAL");
  if (real_arule == affix_rule_nil)
    dcg_panic ("could not locate primitive type REAL");
  text_arule = lookup_affix_tree (predef -> adefs, "TEXT");
  if (text_arule == affix_rule_nil)
    dcg_panic ("could not locate primitive type TEXT");
}

/*
   Check if the root module really exports affix rules
*/
static void determine_real_exports ()
{ export_list exports = root_module -> Root_module.exports;
  int ix;
  if (!partial_compilation) return;
  for (ix = 0; ix < exports -> size; ix++)
    { export exp = exports -> array[ix];
      affix_rule arule;
      if (exp -> tag != TAGExport_affix) continue;
      arule = exp -> Export_affix.adef;
      if (arule == NULL)
        dcg_bug ("determine_real_exports",
		 "exported affix rule '%s' should have been identified", arule -> aname);
      arule = get_repr_affix_rule (arule);
      arule -> exported = 1;
    };
}

/*
   Try dump all affix rules
*/
static void try_dump_affix_rules ()
{ int ix, iy;
  if (!dump_affix_rules) return;
  for (ix = 0; ix < all_affix_rules -> size; ix++)
    { affix_rule arule = all_affix_rules -> array[ix];
      int_list lc_nrs = arule -> lc_nrs;
      dcg_eprint ("\t %d: %s", ix, arule -> aname);
      if (lc_nrs != int_list_nil)
	{ dcg_eprint ("\tLC nrs:");
	  for (iy = 0; iy < lc_nrs -> size; iy++)
	    dcg_eprint ("%s %d", (iy)?",":"", lc_nrs -> array[iy]);
	};
      dcg_wlog ("");
    };
}

/*
   The driver for the affix rules check across all modules
   When no errors have been found, we are also sure (if not panic)
   that the primitive types exist and have been imported in all modules
   Since they are going to be needed in the subsequent type check
   of the procedures (rules), we preallocate them to avoid constantly
   looking for the locally imported one.
*/
void check_affix_rules ()
{ dcg_warning (0, "   checking affix rules...");
  register_and_export_affix_rules ();
  dcg_hint ("      collected %d affix rules", all_affix_rules -> size);
  import_and_identify_affix_rules (); 
  dcg_panic_if_errors ();
  detect_chain_productions ();
  dcg_panic_if_errors ();
  resolve_synonym_rules ();
  check_wellformedness ();
  dcg_panic_if_errors ();
  determine_leftcorner_relation ();
  create_primitive_type_refs ();
  determine_representations ();
  determine_real_exports ();
  try_dump_affix_rules ();
}

/*
   Primitive types are INT, REAL or TEXT
*/
static int is_primitive_type (affix_rule arule, string tname)
{ affix_rule rhs_rule;
  if ((arule == affix_rule_nil) || (arule == AFFIX_ERROR)) return (0);
  rhs_rule = get_prime_affix_rule (arule);
  return ((rhs_rule -> tag == TAGAffix_prim) && streq (rhs_rule -> aname, tname));
}

affix_rule get_primitive_type (prim_type pt)
{ switch (pt)
    { case prim_int:  return (int_arule);
      case prim_real: return (real_arule);
      case prim_text: return (text_arule);
      default: dcg_bad_tag (pt, "get_primitive_type");
    };
  return (affix_rule_nil);
}

/*
   Primitive type recognition
*/
int is_a_primitive_type (affix_rule arule)
{ affix_rule rhs_rule;
  if ((arule == affix_rule_nil) || (arule == AFFIX_ERROR)) return (0);
  rhs_rule = get_prime_affix_rule (arule);
  return (rhs_rule -> tag == TAGAffix_prim);
}

int is_integer_type (affix_rule arule)
{ return (is_primitive_type (arule, "INT"));
}

int is_real_type (affix_rule arule)
{ return (is_primitive_type (arule, "REAL"));
}

int is_text_type (affix_rule arule)
{ return (is_primitive_type (arule, "TEXT"));
}

/*
   When the affix rule system is checked, we can identify affix variables
   Perhaps MAXBUFLEN should move to some general header file:
   the lexer.c needs it as well.
*/
#define MAXBUFLEN 1024
affix_rule identify_affix_instance (string vname, module mod)
{ char my_buf[MAXBUFLEN];
  int len = strlen (vname);
  char *ptr = my_buf + len - 1;
  affix_rule arule;
  strcpy (my_buf, vname);
  while (('0' <= *ptr) && (*ptr <= '9')) ptr--;
  ptr++;
  *ptr = '\0';
  arule = lookup_affix_tree (mod -> adefs, my_buf);
  return (get_prime_affix_rule (arule));
}
