/*
   File: coder.c
   Generates the final C code

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: coder.c,v 1.25 2012/01/20 09:09:20 marcs Exp $"
*/

/* include config.h if autoconfigured */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* standard includes */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#endif

/* include sys/time.h with or without time.h */
#if TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   if HAVE_SYS_TIME_H
#      include <sys/time.h>
#   else
#      include <time.h>
#   endif
#endif

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "cdl_ds.h"
#include "options.h"
#include "ast_utils.h"
#include "globals.h"
#include "affix_rules.h"
#include "modules.h"
#include "coder.h"

/* TMPDIR is the usual environment variable to consult */
#ifndef TMPDIR
#define TMPDIR "TMPDIR"
#endif

/* Define the usual default temporary directory */
#ifndef DEFAULT_TMP_DIR
#define DEFAULT_TMP_DIR "/tmp"
#endif

/* The C output */
static FILE *code;
static FILE *header;
static void try_open_output_files ()
{ header = NULL;
  if (!generate_c_code)
    { /* We generate the c_code in a temporary location */
      char *temp_dir = getenv (TMPDIR);
      if (temp_dir == NULL) temp_dir = DEFAULT_TMP_DIR;
      c_code_fname = dcg_new_fmtd_string ("%s%cmcdl_%d.c", temp_dir, DIR_SEP, (int) getpid ());
      header_fname = NULL;
    }
  else
    { /* Generate C code in situ. Note partial compilation implies -C */
      if (output_fname != NULL) c_code_fname = output_fname;
      else c_code_fname = dcg_new_fmtd_string ("%s.c", base_mname);
      if (partial_compilation && (header_fname == NULL))
        header_fname = dcg_new_fmtd_string ("%s.h", base_mname);
      else header_fname = NULL;
    };

  code = fopen (c_code_fname, "w");
  if (code == NULL)
    dcg_panic ("could not open code file %s", c_code_fname);
  if (!partial_compilation) return;
  header = fopen (header_fname, "w");
  if (header == NULL)
    dcg_panic ("could not open header file %s", header_fname);
}

static void code_string (char *format, ...)
{ char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  fprintf (code, "%s", buf);
};

static void code_line (char *format, ...)
{ char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  fprintf (code, "%s\n", buf);
};

static void header_string (char *format, ...)
{ char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  fprintf (header, "%s", buf);
};

static void header_line (char *format, ...)
{ char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  fprintf (header, "%s\n", buf);
};

static void code_file_headers ()
{ time_t thetime;
  char *atime;
  time (&thetime);
  atime = ctime (&thetime);
  code_line ("/*");
  code_line ("   File: %s", c_code_fname);
  code_string ("   Generated on %s", atime);
  code_line ("*/\n");
  if (partial_compilation)
    { header_line ("/*");
      header_line ("   File: %s", header_fname);
      header_string ("   Generated on %s", atime);
      header_line ("*/");
      header_line ("#ifndef IncMcdl_%s", base_mname);
      header_line ("#define IncMcdl_%s", base_mname);
      header_line ("\n/* Runtime system includes */");
      header_line ("#include <mrts_mcdl.h>");
      header_line ("#include <mrts_stdtypes.h>");
      header_line ("#include <mrts_predef.h>\n");
      header_line ("/* For compiling with C++ */");
      header_line ("#ifdef __cplusplus");
      header_line ("extern \"C\" {");
      header_line ("#endif /* __cplusplus */\n");
      header_line ("/* CDL3 code initialization */");
      header_line ("void M%s_INIT (int argc, char **argv);\n", root_module -> mname);
      header_line ("/* Exported algorithms and affix rules */");
    };
  code_line ("/* Standard includes */");
  code_line ("#include <stdio.h>");
  code_line ("#include <stdlib.h>");
  code_line ("#include <string.h>\n");
  code_line ("/* Runtime system includes */");
  code_line ("#include <mrts_mcdl.h>");
  code_line ("#include <mrts_stdtypes.h>");
  code_line ("#include <mrts_predef.h>");
  if (partial_compilation)
    code_line ("#include \"%s\"", header_fname);
}

/*
   Coding of type system
*/
static void code_forward_rts_decls ()
{ int ix;
  for (ix = 0; ix < all_affix_rules -> size; ix++)
    { affix_rule arule = all_affix_rules -> array[ix];
      int exported = arule -> exported;
      if (arule -> tag == TAGAffix_alts)
	{ if (exported)
	    { header_line ("\n/* Declare exported type ops for affix rule %s */", arule -> aname);
	      header_line ("int  Equal_T%d (value v1, value v2);", arule -> anr);
	      header_line ("void Write_T%d (value v);", arule -> anr);
	      header_line ("#define Equal_%s Equal_T%d", arule -> aname, arule -> anr);
	      header_line ("#define Write_%s Write_T%d\n", arule -> aname, arule -> anr);
	    }
	  else
	    { code_line ("\n/* Forward declare type ops for affix rule %s */", arule -> aname);
	      code_line ("static int  Equal_T%d (value v1, value v2);", arule -> anr);
	      code_line ("static void Write_T%d (value v);", arule -> anr);
	    };
	};
      if (!arule -> needs_gc) continue;
      if (arule -> tag == TAGAffix_alts)
	{ if (exported)
	    { header_line ("void Detach_T%d (value v);", arule -> anr);
	      header_line ("#define Attach_T%d(v) Attach((v))", arule -> anr);
	      header_line ("#define Detach_%s Detach_T%d", arule -> aname, arule -> anr);
	      header_line ("#define Attach_%s(v) Attach((v))\n", arule -> aname);
	    }
          else
	    { code_line ("static void Detach_T%d (value v);", arule -> anr);
	      code_line ("#define     Attach_T%d(v) Attach((v))", arule -> anr);
	    };
	};
    }
}

static void code_equal_elems (affix_element_list elems)
{ int partnr = 0;
  int ix;
  for (ix = 0; ix < elems -> size; ix++)
    { affix_element elem = elems -> array[ix];
      affix_rule vdef;
      if (elem -> tag != TAGAffix_var) continue;
      vdef = get_repr_affix_rule (elem -> Affix_var.vdef);
      switch (vdef -> tag)
	{ case TAGAffix_prim:
	    code_line ("\tif (!Equal_%s (PartValue (v1, %d), PartValue (v2, %d))) return (0);",
		       vdef -> aname, partnr + 1, partnr + 1);
	    break;
	  case TAGAffix_alts:
	    code_line ("\tif (!Equal_T%d (PartValue (v1, %d), PartValue (v2, %d))) return (0);",
		       vdef -> anr, partnr + 1, partnr + 1);
	    break;
	  default: dcg_bad_tag (elem -> tag, "code_equal_elems");
	};
      partnr++;
    };
  code_line ("\tbreak;");
}

static void code_equality_test (affix_rule arule)
{ affix_alternative_list alts = arule -> Affix_alts.alts;
  int exported = arule -> exported;
  int ix;
  code_line ("\n/* Equality test for affix rule %s */", arule -> aname);
  code_line ("%sint Equal_T%d (value v1, value v2)", (exported)?"":"static ", arule -> anr);
  if (!arule -> Affix_alts.enumerable)
    { code_line ("{ if (Vptr (v1) == Vptr (v2)) return (1);");
      code_line ("  if (AlternativeNr (v1) != AlternativeNr (v2)) return (0);");
      code_line ("  switch (AlternativeNr (v1))");
      for (ix = 0; ix < alts -> size; ix++)
        { code_line ("    %s case %d:", (ix == 0)?"{":" ", ix);
          code_equal_elems (alts -> array[ix] -> elems);
        };
      code_line ("      default: mrts_panic (\"bad tag %%d in Equal_T%d\", AlternativeNr (v1));",
    	     arule -> anr);
      code_line ("    };");
      code_line ("  return (1);");
    }
  else code_line ("{ return (Value_INT (v1) == Value_INT (v2));");
  code_line ("}");
}

static void code_write_elems (affix_element_list elems, int alt_nr)
{ int partnr = 0;
  int ix;
  code_line ("\tfputs (\"\\n( Tag %d\", stderr);", alt_nr);
  for (ix = 0; ix < elems -> size; ix++)
    { affix_element elem = elems -> array[ix];
      code_line ("\tfputs (\", \", stderr);");
      switch (elem -> tag)
 	{ case TAGAffix_term:
	    code_line ("\tfputs (\"%s\", stderr);", elem -> Affix_term.tname);
	    break; 
	  case TAGAffix_var:
	    { affix_rule vdef = get_repr_affix_rule (elem -> Affix_var.vdef);
	      switch (vdef -> tag)
		{ case TAGAffix_prim:
		    code_line ("\tWrite_%s (PartValue (v, %d));", vdef -> aname, partnr + 1);
		    break;
		  case TAGAffix_alts:
		    code_line ("\tWrite_T%d (PartValue (v, %d));", vdef -> anr, partnr + 1);
		    break;
		  default: dcg_bad_tag (elem -> tag, "code_write_elems");
		};
	      partnr++;
	    };
	    break;
	  default: dcg_bad_tag (elem -> tag, "code_write_elems");
	};
    };
  code_line ("\tfputs (\")\", stderr);");
  code_line ("\tbreak;");
}

static void code_write_routine (affix_rule arule)
{ affix_alternative_list alts = arule -> Affix_alts.alts;
  int exported = arule -> exported;
  int ix;
  code_line ("\n/* Default Write routine for affix_rule %s */", arule -> aname);
  code_line ("%svoid Write_T%d (value v)", (exported)?"":"static ", arule -> anr);
  if (arule -> Affix_alts.enumerable) code_line ("{ switch (Value_INT (v))");
  else code_line ("{ switch (AlternativeNr (v))");
  for (ix = 0; ix < alts -> size; ix++)
    { code_line ("    %s case %d:", (ix == 0)?"{":" ", ix);
      code_write_elems (alts -> array[ix] -> elems, ix);
    };
  if (arule -> Affix_alts.enumerable)
    code_line ("      default: mrts_panic (\"bad tag %%d in Write_T%d\", Value_INT (v));",
	       arule -> anr);
  else code_line ("      default: mrts_panic (\"bad tag %%d in Write_T%d\", AlternativeNr (v));",
		  arule -> anr);
  code_line ("    };");
  code_line ("}");
}

static void code_detach_elems (affix_element_list elems)
{ int partnr = 0;
  int ix;
  for (ix = 0; ix < elems -> size; ix++)
    { affix_element elem = elems -> array[ix];
      affix_rule vdef;
      if (elem -> tag != TAGAffix_var) continue;
      vdef = elem -> Affix_var.vdef;
      if (vdef -> needs_gc)
	{ vdef = get_repr_affix_rule (vdef);
	  switch (vdef -> tag)
	    { case TAGAffix_prim:
		code_line ("\tDetach_%s (PartValue (v, %d));", vdef -> aname, partnr + 1);
		break;
	      case TAGAffix_alts:
		code_line ("\tDetach_T%d (PartValue (v, %d));", vdef -> anr, partnr + 1);
		break;
	      default: dcg_bad_tag (vdef -> tag, "code_detach_elems");
	    };
	};
      partnr++;
    };
  code_line ("\tFree (v, %d);", partnr + 1);
  code_line ("\tbreak;");
}

static void code_detach_routine (affix_rule arule)
{ affix_alternative_list alts = arule -> Affix_alts.alts;
  int exported = arule -> exported;
  int ix;
  code_line ("\n/* Detach routine for affix_rule %s */", arule -> aname);
  code_line ("%svoid Detach_T%d (value v)", (exported)?"":"static ", arule -> anr);
  code_line ("{ if (!TryDecrRefcount (v)) return;");
  code_line ("  switch (AlternativeNr (v))");
  for (ix = 0; ix < alts -> size; ix++)
    { code_line ("    %s case %d:", (ix == 0)?"{":" ", ix);
      code_detach_elems (alts -> array[ix] -> elems);
    }
  code_line ("      default: mrts_panic (\"bad tag %%d in Detach_T%d\", AlternativeNr (v));",
	     arule -> anr);
  code_line ("    };");
  code_line ("}");
}

static void code_rts_decls ()
{ int ix;
  for (ix = 0; ix < all_affix_rules -> size; ix++)
    { affix_rule arule = all_affix_rules -> array[ix];
      if (arule -> tag == TAGAffix_alts)
	{ code_equality_test (arule);
	  code_write_routine (arule);
	};
      if (!arule -> needs_gc) continue;
      if (arule -> tag == TAGAffix_alts)
	code_detach_routine (arule);
    }
}

/*
   Coding of the global variables and arrays
*/
static void code_global_variable (variable glb, int ix)
{ switch (glb -> tag)
    { case TAGModule_var:
	code_line ("static value G%d = InitValue;\t\t/* %s */", ix, glb -> vname);
	break;
      case TAGModule_array:
	code_line ("static Array G%d = InitArray;\t\t/* %s */", ix, glb -> vname);
	break;
      default: dcg_bad_tag (glb -> tag, "code_global_variable");
    };
}

static void code_global_variables ()
{ int ix;
  code_line ("\n/* All global variables */");
  for (ix = 0; ix < all_global_variables -> size; ix++)
    code_global_variable (all_global_variables -> array[ix], ix);
}

/*
   Collecting and coding of constants
*/

/* introduce type and nil definitions */
typedef struct constant_rec *constant_tree;
struct constant_rec
        { int tag_key;
	  string text_key;
          int const_nr;
          constant_tree left;
          constant_tree right;
        };
#define constant_tree_nil ((constant_tree) NULL)

static int constant_key_cmp (int tag, string s, constant_tree current)
{ if (tag < current -> tag_key) return (-1);
  if (tag > current -> tag_key) return (1);
  if ((s == string_nil) && (current -> text_key == string_nil)) return (0);
  if (s == string_nil) return (-1);
  if (current -> text_key == string_nil) return (1);
  return (strcmp (s, current -> text_key));
}

/* Adminstration */
static int nr_constants;
static constant_tree all_constants;

int enter_constant (int tag, string s)
{ constant_tree *ptr = &all_constants;
  constant_tree new;
  while (*ptr != constant_tree_nil)
    { constant_tree current = *ptr;
      int cond = constant_key_cmp (tag, s, current);
      if (cond < 0) ptr = &current -> left;
      else if (cond > 0) ptr = &current -> right;
      else /* Found constant, return nr */
        return (current -> const_nr);
    };

  new = (constant_tree) dcg_malloc (sizeof (struct constant_rec));
  new -> tag_key = tag;
  new -> text_key = (s == string_nil)?string_nil:attach_string (s);
  new -> const_nr = nr_constants;
  new -> left = constant_tree_nil;
  new -> right = constant_tree_nil;
  nr_constants++;
  *ptr = new;
  return (new -> const_nr);
};

void code_text_denotation (string tag)
{ char *ptr = tag;
  for (; *ptr; ptr++)
    { char ch = *ptr;
      switch (ch)
	{ case '\a': code_string ("\\a"); break;
	  case '\b': code_string ("\\b"); break;
	  case '\f': code_string ("\\f"); break;
	  case '\n': code_string ("\\n"); break;
	  case '\r': code_string ("\\r"); break;
	  case '\t': code_string ("\\t"); break;
	  case '\v': code_string ("\\v"); break;
	  case '"' : code_string ("\\\""); break;
	  case '\\': code_string ("\\\\"); break;
	  default:
	    if ((unsigned char) ch < ' ') code_string ("\\%03o", ch);
	    else code_string ("%c", ch);
	};
    };
}

/*
   Coding of constants
*/
void code_constant_tree (constant_tree tree)
{ if (tree == constant_tree_nil) return;
  code_constant_tree (tree -> left);
  code_string (" { %d, %d, ", tree -> const_nr, tree -> tag_key);
  if (tree -> text_key == string_nil) code_line ("NULL },");
  else
    { code_string ("\"");
      code_text_denotation (tree -> text_key);
      code_line ("\" },");
    };
  code_constant_tree (tree -> right);
}

void code_constants ()
{ code_line ("\n/* Constants */");
  code_line ("static const_init const_init_table[%d] =", nr_constants);
  code_line ("{");
  code_constant_tree (all_constants);
  code_line ("};\n");
  code_line ("static value Const[%d];", nr_constants);
}

/*
   Coding of procedure names and arguments
*/
static void gen_fpars_names (FILE *f, fpar_list fpars)
{ int ix;
  for (ix = 0; ix < fpars -> size; ix++)
    fprintf (f, "_%s", fpars -> array[ix] -> fdef -> adef -> aname);
}

static void gen_proc_name (FILE *f, proc pdef)
{ switch (pdef -> tag)
    { case TAGExt_proc: fprintf (f, "%s", pdef -> Ext_proc.ename); break;
      case TAGAlts:
        { string mname = all_modules -> array[pdef -> mnr] -> mname;
          fprintf (f, "M%s_P%s", mname, pdef -> pspec -> concat_name);
	}; break;
      case TAGRoot:
	{ string mname = all_modules -> array[pdef -> mnr] -> mname;
          fprintf (f, "M%s_%s", mname, pdef -> pspec -> concat_name);
	}; break;
      default: dcg_bad_tag (pdef -> tag, "gen_proc_name");
    };
  gen_fpars_names (f, pdef -> pspec -> p1fpars);
}

static void gen_proc_fpars (FILE *f, proc pdef)
{ fpar_list fpars = pdef -> pspec -> p1fpars;
  int ix;
  fprintf (f, " (");
  for (ix = 0; ix < fpars -> size; ix++)
    { fpar fp = fpars -> array[ix];
      fprintf (f, "%svalue %sP%d", (ix)?", ":"", (fp -> fdir == d_inherited)?"":"*", ix);
    };
  fprintf (f, ")");
}

static void code_forward_procs ()
{ module predef = identify_module (PREDEF_NAME);
  int ix;
  code_line ("\n/* Forward declare all used procedures */");
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc pdef = all_procedures -> array[ix];
      if (!pdef -> called) continue;
      if (pdef -> mnr == predef -> mnr) continue;
      switch (pdef -> tag)
        { case TAGExt_proc: code_string ("int "); break;
          case TAGRoot:
	    if (partial_compilation)
	      { header_string ("void ");
		gen_proc_name (header, pdef);
	        gen_proc_fpars (header, pdef);	/* should be none */
	        header_line (";");
		code_string ("void ");
	      }
	    else code_string ("static void ");
	    break;
	  case TAGAlts:
	    if (pdef -> Alts.exported)
	      { header_string ("int ");
		gen_proc_name (header, pdef);
	        gen_proc_fpars (header, pdef);
	        header_line (";");
		code_string ("int ");
	      }
	    else code_string ("static int ");
	  default: break;
	};
      
      gen_proc_name (code, pdef);
      gen_proc_fpars (code, pdef);
      code_line (";");
    };
}

/*
   Coding of local variables
*/
static void code_locals (proc pdef)
{ int nr_fpars = pdef -> pspec -> p1fpars -> size;
  int ix;
  for (ix = 0; ix < pdef -> pvars -> size; ix++)
    { variable pvar = pdef -> pvars -> array[ix];
      code_string ("  value L%d; /* %s", ix, pvar -> vname);
      if (ix < nr_fpars) code_string (", local copy");
      code_line (" */");
    };
  for (ix = 0; ix < pdef -> ivars; ix++)
    code_line ("  value *I%d;", ix);
}

/*
   Coding of intermediate code
*/
static void code_address (address adr)
{ switch (adr -> tag)
    { case TAGParameter: code_string ("P%d", adr -> Parameter.pnr); break;
      case TAGGlobal:	 code_string ("G%d", adr -> Global.gnr); break;
      case TAGLocal:	 code_string ("L%d", adr -> Local.lnr); break;
      case TAGIlocal:	 code_string ("I%d", adr -> Ilocal.inr); break;
      case TAGTconst:	 code_string ("Const[%d]", adr -> Tconst.cnr); break;
      case TAGIndirect:	
	code_string ("*");
	code_address (adr -> Indirect.adr);
	break;
      case TAGReference:
	code_string ("&");
	code_address (adr -> Reference.adr);
	break;
      case TAGPart:
	code_string ("PartValue(");
	code_address (adr -> Part.adr);
	code_string (", %d)", adr -> Part.idx);
	break;
      default: dcg_bad_tag (adr -> tag, "code_address");
    };
}

static void code_address_list (address_list adrs)
{ int ix;
  code_string (" (");
  for (ix = 0; ix < adrs -> size; ix++)
    { if (ix) code_string (", ");
      code_address (adrs -> array[ix]);
    };
  code_string (")");
}

static void code_operator (operator op)
{ switch (op)
    { case times:		code_string ("*"); break;
      case modulo:		code_string ("%%"); break;
      case divides:		code_string ("/"); break;
      case plus:		code_string ("+"); break;
      case minus:		code_string ("-"); break;
      case shift_left:		code_string ("<<"); break;
      case shift_right:		code_string (">>"); break;
      case bitwise_and:		code_string ("&"); break;
      case bitwise_or:		code_string ("|"); break;
      case bitwise_xor:		code_string ("^"); break;
      case bitwise_not:		code_string ("~"); break;
      case int_times_int:	code_string ("*"); break;
      case real_times_real:	code_string ("*"); break;
      case int_times_text:	code_string ("*"); break;
      case int_plus_int:	code_string ("+"); break;
      case real_plus_real:	code_string ("+"); break;
      case text_plus_text:	code_string ("+"); break;
      case int_minus_int:	code_string ("-"); break;
      case real_minus_real:	code_string ("-"); break;
      default: dcg_bad_tag (op, "dump_operator");
    };
}

static void code_iexpr (iexpr exp)
{ switch (exp -> tag)
    { case TAGIvref:	 
	code_string ("Value_INT (");
	code_address (exp -> Ivref.adr);
	code_string (")");
	break;
      case TAGIdyop:
	code_string ("(");
	code_iexpr (exp -> Idyop.src1);
	code_string (" ");
	code_operator (exp -> Idyop.dop);
	code_string (" ");
	code_iexpr (exp -> Idyop.src2);
	code_string (")");
	break;
      case TAGImonop:
	code_string ("(");
	code_operator (exp -> Imonop.mop);
	code_iexpr (exp -> Imonop.src);
	code_string (")");
	break;
      case TAGInum:
	code_string ("%dL", exp -> Inum.num);
	break;
      default: dcg_bad_tag (exp -> tag, "code_iexpr");
    };
}

static void code_rexpr (rexpr exp)
{ switch (exp -> tag)
    { case TAGRvref:	 
	code_string ("Value_REAL (");
	code_address (exp -> Rvref.adr);
	code_string (")");
	break;
      case TAGRdyop:
	code_string ("(");
	code_rexpr (exp -> Rdyop.src1);
	code_string (" ");
	code_operator (exp -> Rdyop.dop);
	code_string (" ");
	code_rexpr (exp -> Rdyop.src2);
	code_string (")");
	break;
      case TAGRmonop:
	code_string ("(");
	code_operator (exp -> Rmonop.mop);
	code_rexpr (exp -> Rmonop.src);
	code_string (")");
	break;
      case TAGRnum:
	code_string ("%g", exp -> Rnum.num);
	break;
      default: dcg_bad_tag (exp -> tag, "code_rexpr");
    };
}

static void code_cases_list (altnr_label_pair_list cases, int fail)
{ int ix;
  code_string ("    { ");
  for (ix = 0; ix < cases -> size; ix++)
    { altnr_label_pair cp = cases -> array[ix];
      if (ix) code_string ("      ");
      code_line ("case %d: goto LAB%d;", cp -> altnr, cp -> lnr);
    };
  code_line ("      default: goto LAB%d;", fail);
  code_string ("    };");
}

static void code_instruction (instr ins)
{ if (ins -> dont_code)
    code_string ("  /* Opt: ");
  switch (ins -> tag)
    { case TAGNop:
      case TAGProc_entry:	break;
      case TAGLabel:		code_string ("LAB%d:", ins -> Label.lnr); break;
      case TAGComment:		code_string ("  /* %s */", ins -> Comment.text); break;
      case TAGSucceed:		code_string ("  return (1);"); break;
      case TAGExit:		code_string ("  mrts_exit ();"); /* fall thru, keep gcc happy */
      case TAGFail:		code_string ("  return (0);"); break;
      case TAGAbort:
	code_line ("  mrts_abort (\"%s\", %d);",
		   all_modules -> array[ins -> Abort.mnr] -> mname, ins -> Abort.lnr);
	code_string ("  return (0);");
	break;
      case TAGAttach:
	{ affix_rule arule = get_repr_affix_rule (all_affix_rules -> array[ins -> Attach.tnr]);
	  if (arule -> tag == TAGAffix_prim)
	    code_string ("  Attach_%s (", arule -> aname);
	  else code_string ("  Attach_T%d (", arule -> anr);
	  code_address (ins -> Attach.loc);
	  code_string (");");
	}; break;
      case TAGDetach:
	{ affix_rule arule = get_repr_affix_rule (all_affix_rules -> array[ins -> Detach.tnr]);
	  if (arule -> tag == TAGAffix_prim)
	    code_string ("  Detach_%s (", arule -> aname);
	  else code_string ("  Detach_T%d (", arule -> anr);
	  code_address (ins -> Detach.loc);
	  code_string (");");
	}; break;
      case TAGCreate:
	code_string ("  ");
	code_address (ins -> Create.dst);
	code_string (" = mrts_create_value (%d,%d);", ins -> Create.altnr, ins -> Create.size);
	break;
      case TAGEqual_altnr:
	code_string ("  if (AlternativeNr (");
	code_address (ins -> Equal_altnr.src);
	code_string (") != %d) goto LAB%d;", ins -> Equal_altnr.altnr, ins -> Equal_altnr.fail);
	break;
      case TAGEqual:
	{ affix_rule arule = get_repr_affix_rule (all_affix_rules -> array[ins -> Equal.tnr]);
	  if (arule -> tag == TAGAffix_prim)
	    code_string ("  if (!Equal_%s (", arule -> aname);
	  else code_string ("  if (!Equal_T%d (", arule -> anr);
	  code_address (ins -> Equal.src1);
	  code_string (", ");
	  code_address (ins -> Equal.src2);
	  code_string (")) goto LAB%d;", ins -> Equal.fail);
	}; break;
      case TAGEqual_int:
	code_string ("  if (Value_INT (");
	code_address (ins -> Equal_int.src);
	code_string (") != (");
	code_iexpr (ins -> Equal_int.cmp);
	code_string (")) goto LAB%d;", ins -> Equal_int.fail);
	break;
      case TAGEqual_real:
	code_string ("  if (Value_REAL (");
	code_address (ins -> Equal_real.src);
	code_string (") != (");
	code_rexpr (ins -> Equal_real.cmp);
	code_string (")) goto LAB%d;", ins -> Equal_real.fail);
	break;
      case TAGCopy:
	code_string ("  ");
	code_address (ins -> Copy.dst);
	code_string (" = ");
	code_address (ins -> Copy.src);
	code_string (";");
	break;
      case TAGCopy_int:
	code_string ("  ");
	code_address (ins -> Copy_int.dst);
	code_string (" = Create_INT (");
	code_iexpr (ins -> Copy_int.src);
	code_string (");");
	break;
      case TAGCopy_real:
	code_string ("  ");
	code_address (ins -> Copy_real.dst);
	code_string (" = Create_REAL (");
	code_rexpr (ins -> Copy_real.src);
	code_string (");");
	break;
      case TAGCallp:
	code_string ("  (void) ");
	gen_proc_name (code, all_procedures -> array[ins -> Callp.pnr]);
	code_address_list (ins -> Callp.args);
	code_string (";");
	break;
      case TAGCallf:
	code_string ("  if (!");
	gen_proc_name (code, all_procedures -> array[ins -> Callf.pnr]);
	code_address_list (ins -> Callp.args);
	code_string (") goto LAB%d;", ins -> Callf.fail);
	break;
      case TAGIndex:
	code_string ("  ");
	code_address (ins -> Index.iloc);
	code_string (" = mrts_index_array (&");
	code_address (ins -> Index.glb);
	code_string (", ");
	code_iexpr (ins -> Index.idx);
	code_string (");");
	break;
      case TAGConc_text:
	{ address_list srcs = ins -> Conc_text.srcs;
	  int ix;
	  for (ix = 0; ix < srcs -> size; ix++)
	    { code_string ("  concat_array[%d] = ", ix);
	      code_address (srcs -> array[ix]);
	      code_line (";");
	    };
	  code_string ("  ");
	  code_address (ins -> Conc_text.dst);
	  code_string (" = mrts_concat_texts (concat_array, %d);", srcs -> size);
	}; break;
      case TAGRepl_text:
	code_string ("  ");
        code_address (ins -> Repl_text.dst);
	code_string (" = mrts_repl_text (");
	code_iexpr (ins -> Repl_text.fac);
	code_string (", ");
	code_address (ins -> Repl_text.src);
	code_string (");");
	break;
      case TAGSwitch_altnr:
	code_string ("  switch (AlternativeNr (");
	code_address (ins -> Switch_altnr.src);
	code_line ("))");
	code_cases_list (ins -> Switch_altnr.cases, ins -> Switch_altnr.fail);
	break;
      case TAGSwitch_int:
	code_string ("  switch (Value_INT (");
	code_address (ins -> Switch_int.src);
	code_line ("))");
	code_cases_list (ins -> Switch_int.cases, ins -> Switch_int.fail);
	break;
      default: dcg_bad_tag (ins -> tag, "code_instruction");
    };
  if (ins -> dont_code) code_line (" */");
  else code_line ("");
}

static void code_basic_block (block bb, int nr)
{ instr ins;
  code_string ("\n  /* Basic block %d", nr);
  if (bb -> dont_code)
    { code_line (", optimized away */");
      return;
    };
  code_line (" */");
  for (ins = bb -> head; ins != instr_nil; ins = ins -> next)
    code_instruction (ins);
}

static void code_procs ()
{ int ix;
  for (ix = 0; ix < all_procedures -> size; ix++)
    { proc pdef = all_procedures -> array[ix];
      spec pspec = pdef -> pspec;
      int iy;
      if (!pdef -> called) continue;
      if (pdef -> tag == TAGExt_proc) continue;
      code_line ("\n/* Code for module %s, %s %s */", all_modules -> array[pdef -> mnr] -> mname,
	         string_from_proc_type (pspec -> ptype), pspec -> canonic_name);
      if (pdef -> tag == TAGRoot) code_string ("%svoid ", (partial_compilation)?"":"static ");
      else code_string ("%sint ", (pdef -> Alts.exported)?"":"static ");
      gen_proc_name (code, pdef);
      gen_proc_fpars (code, pdef);
      code_line ("\n{");
      if (pdef -> concat_size)
	code_line ("  value concat_array[%d];", pdef -> concat_size);
      code_locals (pdef);
      for (iy = 0; iy < pdef -> bbs -> size; iy++)
        code_basic_block (pdef -> bbs -> array[iy], iy);
      code_line ("}");
    }
}

static void code_main ()
{ if (partial_compilation)
    { code_line ("\n/* Initialization code */");
      code_line ("void M%s_INIT (int argc, char **argv)", root_module -> mname);
    }
  else
    { code_line ("\n/* Main program */");
      code_line ("int main (int argc, char **argv)");
    };
  code_line ("{ mrts_init_basic ();");
  code_line ("  mrts_init_predef (argc, argv);");
  code_line ("  mrts_init_constants (%d, const_init_table, Const);", nr_constants);
  if (!partial_compilation)
    { code_line ("  M%s_PRELUDE ();", root_module -> mname);
      code_line ("  M%s_ROOT ();", root_module -> mname);
      code_line ("  M%s_POSTLUDE ();", root_module -> mname);
      code_line ("  return (0);");
    };
  code_line ("}");
}

static void code_file_trailers ()
{ if (partial_compilation)
    { header_line ("\n#ifdef __cplusplus");
      header_line ("}");
      header_line ("#endif /* __cplusplus */");
      header_line ("#endif /* IncMcdl_%s */", base_mname);
      fclose (header);
    };
  fclose (code);
}

void init_coder ()
{ all_constants = constant_tree_nil;
  nr_constants = 0;
}

void target_code_generation ()
{ if (no_code) return;
  dcg_warning (0, "   generating C code...");
  try_open_output_files ();
  code_file_headers ();
  code_forward_rts_decls ();
  code_rts_decls ();
  code_global_variables ();
  code_constants ();
  code_forward_procs ();
  code_procs ();
  code_main ();
  code_file_trailers ();
};
