/*
   File: contfree.c
   Defines the context free analysis

   This module is also responsible for parsing the modules used
   by the source module or root module in case of integral coding

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: contfree.c,v 1.18 2012/04/21 14:00:00 marcs Exp $"
*/

/* global includes */
#include <stdio.h>   
#include <stdlib.h>
#include <string.h>

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>
#include <dcg_string.h>   
#include <dcg_plist.h>
#include <dcg_plist_ops.h>
               
/* local includes */ 
#include "cdl_ds.h"
#include "affix_rules.h"
#include "options.h"
#include "lexer.h"
#include "parser.h"
#include "modules.h"
#include "contfree.h"

/*
   Locate the source file;
   If found, initialize the lexer
*/
static int try_locate (char *fname)
{ FILE *fd = fopen (fname, "r");
  if (fd == NULL) return (0);
  init_lexer (fd, fname);
  return (1);
}

static int locate_and_init_module (string dir, string sname)
{ char buf[MAXPATHLEN+1];
  if (streq (dir, ".")) sprintf (buf, "%s", sname);
  else sprintf (buf, "%s%c%s", dir, DIR_SEP, sname);
  if (try_locate (buf)) return (1);
  if (streq (dir, ".")) sprintf (buf, "%s.cdl3", sname);
  else sprintf (buf, "%s%c%s.cdl3", dir, DIR_SEP, sname);
  if (try_locate (buf)) return (1);
  if (streq (dir, ".")) sprintf (buf, "%s.cdl", sname);
  else sprintf (buf, "%s%c%s.cdl", dir, DIR_SEP, sname);
  if (try_locate (buf)) return (1);
  if (streq (dir, ".")) sprintf (buf, "%s.k3", sname);
  else sprintf (buf, "%s%c%s.k3", dir, DIR_SEP, sname);
  return (try_locate (buf));
}

static void parse_main_module (module *mod)
{ dcg_hint ("      parsing module %s", base_mname);
  if (!locate_and_init_module (dir_name, source_name))
    dcg_panic ("could not open source file '%s'", source_name);
  should_be_module (base_mname, mod);
  finish_lexer ();
}

static void parse_module (string mname, module *mod)
{ int ix;
  dcg_hint ("      parsing module %s", mname);
  for (ix = 0; ix < module_search_path -> size; ix++)
    if (locate_and_init_module (module_search_path -> array[ix], mname))
      { should_be_module (mname, mod);
        finish_lexer ();
	return;
      };
  dcg_panic ("could not locate module '%s'", mname);
}

static void add_uses_to_module_names (string_list uses, string_list mnames)
{ int ix;
  for (ix = 0; ix < uses -> size; ix++)
    { string mnm = uses -> array[ix];
      add_uniquely_to_string_list (mnames, mnm);
    };
}

/*
   The actual context free analysis
*/
void context_free_analysis ()
{ string_list mnames = new_string_list ();
  string predef_name = new_string (PREDEF_NAME);
  module mod;
  int ix = 1;

  /* parse the main module */
  dcg_warning (0, "   parsing modules...");
  parse_main_module (&mod);
  if (parse_only && dump_parser)
    { dcg_wlog ("   Dump of module parse tree:");
      pp_module (dcg_error_file (), mod);
      dcg_wlog ("");
    };
  dcg_panic_if_errors ();

  /* if we're only parsing, we stop now */
  if (parse_only) exit (0);

  /*
     We have succesfully parsed the first source module
     Register our name and add the predefines to our uses
  */
  register_module (mod);
  app_string_list (mnames, mod -> mname); 
  add_uniquely_to_string_list (mod -> uses, predef_name);
  add_uses_to_module_names (mod -> uses, mnames);

  /*
     Read the modules on the module name list as well as the closure of it:
     we need all the modules it may reference to setup the typing system
  */
  while (ix < mnames -> size)
    { string mname = mnames -> array[ix];
      parse_module (mname, &mod);
      register_module (mod);
      dcg_panic_if_errors ();

      /* Add predef to the uses list unless we are predef itself */
      if (!streq (mname, predef_name))
        add_uniquely_to_string_list (mod -> uses, predef_name);
      add_uses_to_module_names (mod -> uses, mnames);
      ix++;
    };

  if (dump_parser)
    dump_modules ();
}
