/*
   File: members.c
   Checks and resolves guards and calls, simplifies affix expressions
   and infers implicit equals due to consistent substitution

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: members.c,v 1.28 2011/09/02 10:11:23 marcs Exp $"
*/

/* standard includes */
#include <stdio.h>
#include <string.h>

/* support lib includes */
#include <dcg.h>
#include <dcg_alloc.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "cdl_ds.h"
#include "modules.h"
#include "ast_utils.h"
#include "globals.h"
#include "affix_rules.h"
#include "variables.h"
#include "contsens.h"
#include "procedures.h"
#include "members.h"
#include "match.h"
#include "flow.h"

/*
   Simple checking of expressions

   The result of a check is that every affix formula becomes typed
   with the exception of enclosed expressions for which resolution
   might be needed. In effect it is a kind of precomputation that
   is only done once, leaving the type pointer in the formula node.

   Resolution is needed for calls and guards
*/
static void check_apar (apar parm, variable_list pvars, module mod)
{ switch (parm -> tag)
    { case TAGSingle:
	parm -> Single.adef = check_expr (parm -> Single.exp, pvars, mod);
	break;
      case TAGBidir:
        { parm -> Bidir.adef1 = check_expr (parm -> Bidir.exp1, pvars, mod);
	  parm -> Bidir.adef2 = check_expr (parm -> Bidir.exp2, pvars, mod);
	};
	break;
      default: dcg_bad_tag (parm -> tag, "check_apar");
    };
}

static void check_apar_list (apar_list parms, variable_list pvars, module mod)
{ int ix;
  for (ix = 0; ix < parms -> size; ix++)
    check_apar (parms -> array[ix], pvars, mod);
}

/*
   Replace the current guard by a single member derived from type analysis.
   Most arithmetic guards and equal guards are replaced in the AST by this call.
   Note: such a guard does not result in new local variables
*/
static void replace_guard_by_member (member_list mems, int *idx, member new_mem)
{ member m = mems -> array[*idx];
  del_member_list (mems, *idx);
  ins_member_list (mems, *idx, new_mem);
  det_member (&m);
  *idx += 1;
}

/*
   An equal guard must be of the form [ VARA = VARB ]
   where VARA and VARB are of an equivalent type
*/
static void check_equal_guard (member_list mems, int *idx, variable_list pvars, module mod)
{ member m = mems -> array[*idx];
  affix_term_list lhs = m -> Guard.lhs;
  affix_term_list rhs = m -> Guard.rhs;
  affix_term lhs_term = lhs -> array[0];
  affix_term rhs_term = rhs -> array[0];
  affix_rule lhs_rule, rhs_rule;

  /* Check the form of the lhs and rhs */
  if (!is_a_single_affix_variable (lhs) || !is_a_single_affix_variable (rhs))
    { contsens_error (mod, m -> line, m -> col,
		      "An equal guard can only be applied to two variables");
      (void) check_expr (lhs, pvars, mod);
      (void) check_expr (rhs, pvars, mod);
      *idx += 1;
      return;
    };

  /* Check type of both sides */
  lhs_rule = check_affix_variable (lhs_term, pvars, mod);
  rhs_rule = check_affix_variable (rhs_term, pvars, mod);
  if (equivalent_affix_rule (lhs_rule, rhs_rule))
    { /* Construct resolved equal, and replace old member in AST */
      member new_equal = new_Res_equal (m -> line, m -> col, attach_affix_term (lhs_term),
				        attach_affix_term (rhs_term));
      replace_guard_by_member (mems, idx, new_equal);
      return;
    };

  /* Complain */
  contsens_error (mod, m -> line, m -> col,
		  "The two sides of the equal guard have different types");
  *idx += 1;
}

/*
   Replace the current guard by the structure derived by analysing
   the sentential forms in the guard. This may result in more guards
   then originally intended by the user. Too bad. If he starts writing
   ambiguous stuff, we sometimes loose a variable here and there
   without recovering memory. Nobis sunt culpa. Also add newly
   created local variables to the local procedure variables
*/
static void add_to_variable_list (variable_list pvars, variable var)
{ var -> nr = pvars -> size;
  app_variable_list (pvars, var);
}

static void try_add_to_variable_list (variable_list pvars, variable var)
{ int ix;
  for (ix = 0; ix < pvars -> size; ix++)
    if (pvars -> array[ix] == var) return;
  add_to_variable_list (pvars, var);
}

static void add_created_local_to_pvars (affix_term term, variable_list pvars)
{ variable var = var_from_affix_term (term);
  if (var -> vname[0] == '_')
    try_add_to_variable_list (pvars, var);
}

static void add_created_locals_to_pvars (member_list structure, variable_list pvars)
{ int ix;
  if (structure == member_list_nil) return;
  for (ix = 0; ix < structure -> size; ix++)
    { member guard = structure -> array[ix];
      switch (guard -> tag)
	{ case TAGAssign:
	    add_created_local_to_pvars (guard -> Assign.src, pvars);
	    add_created_local_to_pvars (guard -> Assign.dst, pvars);
	    break;
	  case TAGRes_guard:
	    add_created_local_to_pvars (guard -> Res_guard.var, pvars);
	  case TAGCast: break;
	  default: dcg_bad_tag (guard -> tag, "add_created_locals_to_pvars");
	};
    };
}

static void insert_guards (member_list mems, int *idx, guard_type type, member_list guards)
{ int ix;
  if (guards == member_list_nil) return;
  if (type == join)
    for (ix = guards -> size - 1; 0 <= ix; ix--)
      ins_member_list (mems, *idx, guards -> array[ix]);
  else
    for (ix = 0; ix < guards -> size; ix++)
      ins_member_list (mems, *idx, guards -> array[ix]);
  *idx += guards -> size;
}

static void update_structure_with_affix_term (guard_type type, affix_term aff_term,
					      member_list structure)
{ while (structure -> size > 0)
    { member last = structure -> array[structure -> size - 1];
      switch (last -> tag)
	{ case TAGRes_guard:
	    { affix_term last_var = last -> Res_guard.var;
	      variable old_local = var_from_affix_term (last_var);
	      last -> Res_guard.var = attach_affix_term (aff_term);
	      detach_variable (&old_local);
	      detach_affix_term (&last_var);
	      return;
	    };
	  case TAGCast:
	    /* Only something to do if it is the first one of the structure */
	    if (structure -> size == 1)
	      { /* Only one affix_term which is abstracted: make an assign */
		/* We need the original type to determine the direction */
	        affix_term source = (type == join)?attach_affix_term (last -> Cast.term):aff_term;
		affix_term target = (type == join)?aff_term:attach_affix_term (last -> Cast.term);
		member assign = new_Assign (last -> line, last -> col, source, target);
		structure -> array[0] = assign;
		detach_member (&last);
	        return;
	      };
	    break;
	  default: dcg_bad_tag (last -> tag, "update_structure_with_affix_term");
	};
      del_member_list (structure, structure -> size - 1);
    };

  dcg_internal_error ("update_structure_with_affix_term");
}

static void replace_guard_by_structure (member_list mems, int *idx, guard_type type,
					affix_term aff_term, member_list structure,
					variable_list pvars)
{ update_structure_with_affix_term (type, aff_term, structure);
  del_member_list (mems, *idx);			/* The old guard */
  insert_guards (mems, idx, type, structure);
  add_created_locals_to_pvars (structure, pvars);
  nonrec_detach_member_list (&structure);
}

/*
   We arrive here when the left hand side of an -> guard is a single variable
   Now we must determine whether we have an assign, arithmetic split, split or join
*/
static void check_undetermined_guard (member_list mems, int *idx, affix_rule lhs_rule,
				      affix_rule rhs_rule, variable_list pvars, module mod)
{ member m = mems -> array[*idx];
  affix_term_list lhs = m -> Guard.lhs;
  affix_term_list rhs = m -> Guard.rhs;
  affix_term lhs_term = lhs -> array[0];
  affix_term rhs_term = rhs -> array[0];
  member_list structure;
  int stat;

  /*
     If guard is of the form LHS -> RHS where LHS and RHS are single
     variables with equivalent types, this is an assign
  */
  if (is_a_single_affix_variable (rhs) && equivalent_affix_rule (lhs_rule, rhs_rule))
    { affix_term dst = rhs -> array[0];
      member new_assign = new_Assign (m -> line, m -> col, attach_affix_term (lhs_term),
				      attach_affix_term (dst));
      replace_guard_by_member (mems, idx, new_assign);
      return;
    };

  /*
     If LHS has a primitive type and the right hand has the same but is some expression,
     we have an arithmetic split (test)
  */
  if (is_a_primitive_type (lhs_rule))
    { /* Check that the right hand side has the same type) */
      if (equivalent_affix_rule (lhs_rule, rhs_rule))
	{ member new_split = new_Res_guard (m -> line, m -> col, arith_split,
					    attach_affix_term (lhs_term), 0, 1,
					    attach_affix_term_list (rhs));
	  replace_guard_by_member (mems, idx, new_split);
	  return;
	}
      else if (rhs_rule == AFFIX_ERROR)
	{ /* error reported at lower level */
	  *idx += 1;
	  return;
	}
      else if (rhs_rule == affix_rule_nil)
	{ contsens_error (mod, m -> line, m -> col, "Affix variable can not be split");
          *idx += 1;
	  return;
	}
    };

  /*
     First try if we can match a split, but if not it may still be a join
     Note that we must ensure that lhs_rule is not primitive
  */
  if (!is_a_primitive_type (lhs_rule))
    { stat = match_terms (mod -> mnr, rhs, lhs_rule, split, &structure);
      if (stat == 1) 
        { /* Successfull match */
          replace_guard_by_structure (mems, idx, split, lhs_term, structure, pvars);
          return;
        }
      else if (stat > 1)
	{ contsens_error (mod, m -> line, m -> col, "This split is ambiguous");
          *idx += 1;
          return;
	}
      else if (!is_a_single_affix_variable (rhs))
	{ contsens_error (mod, m -> line, m -> col, "Affix variable cannot be split");
          *idx += 1;
          return;
	};

      /* Fall thru to join check */
    };

  /*
     We come here if we could not split, perhaps this one is a join
     Note that we must check if the rhs_rule is not primitive
  */
  if (!is_a_primitive_type (rhs_rule) && is_a_single_affix_variable (rhs))
    { stat = match_terms (mod -> mnr, lhs, rhs_rule, join, &structure);
      if (stat == 1)
        { /* Successfull match */
	  replace_guard_by_structure (mems, idx, join, rhs_term, structure, pvars);
	  return;
	}
      else if (stat > 1)
	contsens_error (mod, m -> line, m -> col, "This join is ambiguous");
      else contsens_error (mod, m -> line, m -> col, "Affix variable cannot be split nor joined");
      *idx += 1;
      return;
    }
  else dcg_internal_error ("check_undetermined_guard");

  /* Increment only reached on error */
  contsens_error (mod, m -> line, m -> col, "This guard is not a split, join or assign guard");
  *idx += 1;
}

/*
   In this case, we know that we have a join guard (something of the form 
   LHS -> RHS (single variable). There are still two possibilities, either
   we have an arithmetic join or real join of some tree structure to a type
*/
static void check_join_guard (member_list mems, int *idx, affix_rule lhs_rule, affix_rule rhs_rule,
			      variable_list pvars, module mod)
{ member m = mems -> array[*idx];
  affix_term_list lhs = m -> Guard.lhs;
  affix_term_list rhs = m -> Guard.rhs;
  affix_term rhs_term = rhs -> array[0];
  member_list structure;
  int stat;
  if (is_a_primitive_type (rhs_rule))
    { /* Check that the left hand side has the same type) */
      if (equivalent_affix_rule (lhs_rule, rhs_rule))
	{ member new_join = new_Res_guard (m -> line, m -> col, arith_join,
				           attach_affix_term (rhs_term), 0, 0,
					   attach_affix_term_list (lhs));
	  replace_guard_by_member (mems, idx, new_join);
	  return;
	}
      else if (lhs_rule != AFFIX_ERROR)
	contsens_error (mod, m -> line, m -> col,
		        "The left hand side of this arithmetic join does not have type %s",
			get_affix_rule_name (rhs_rule));
    }
  else
    { /* Check if we can match a join */
      stat = match_terms (mod -> mnr, lhs, rhs_rule, join, &structure);
      if (stat == 1) 
	{ /* Successfull match */
	  replace_guard_by_structure (mems, idx, join, rhs_term, structure, pvars);
	  return;
	}
      else if (stat > 1)
	contsens_error (mod, m -> line, m -> col, "This join is ambiguous");
      else
        contsens_error (mod, m -> line, m -> col, "Affix variable cannot be joined this way");
    };
  *idx += 1;
}

/*
   For the other guards we have to determine if it is an assign, split or join
   guard. When the left side is not a single variable, the guard must be join
*/
static void check_arrow_guard (member_list mems, int *idx, variable_list pvars, module mod)
{ member m = mems -> array[*idx];
  affix_term_list lhs = m -> Guard.lhs;
  affix_term_list rhs = m -> Guard.rhs;
  affix_rule lhs_rule = check_expr (lhs, pvars, mod);
  affix_rule rhs_rule = check_expr (rhs, pvars, mod);
  if (is_a_single_affix_variable (lhs))
    check_undetermined_guard (mems, idx, lhs_rule, rhs_rule, pvars, mod);
  else if (is_a_single_affix_variable (rhs))
    check_join_guard (mems, idx, lhs_rule, rhs_rule, pvars, mod);
  else
    { contsens_error (mod, m -> line, m -> col, "This guard is not a split, join or assign guard");
      *idx += 1;
    }
}
  
static void check_guard (member_list mems, int *idx, variable_list pvars, module mod)
{ if (mems -> array[*idx] -> Guard.typ == equal) check_equal_guard (mems, idx, pvars, mod);
  else check_arrow_guard (mems, idx, pvars, mod);
}

/*
   A call is first preidentified by name and number of first/second pass arguments
*/
static void modify_argument_guards (guard_type type, affix_term *tgt_var, member_list *guards)
{ member_list lguards = *guards;
  member last_guard = lguards -> array[lguards -> size - 1];
  affix_term last_var;
  switch (last_guard -> tag)
    { case TAGRes_guard: last_var = last_guard -> Res_guard.var; break;
      case TAGCast:	 last_var = last_guard -> Cast.term; break;
      default:		 dcg_bad_tag (last_guard -> tag, "modify_argument_guards");
    };

  if (*tgt_var == affix_term_nil)
    *tgt_var = attach_affix_term (last_var);
  else if (last_guard -> tag == TAGRes_guard)
    { variable old_local = var_from_affix_term (last_var);
      last_guard -> Res_guard.var = attach_affix_term (*tgt_var);
      detach_variable (&old_local);
      detach_affix_term (&last_var);
    }
  else
    { affix_term src = (type == join)?last_var:*tgt_var;
      affix_term dst = (type == join)?*tgt_var:last_var;
      member new_assign = new_Assign (last_var -> line, last_var -> col,
				      attach_affix_term (src), attach_affix_term (dst));
      app_member_list (lguards, new_assign);
    };
}

static int ftyp_matches_argument (int curr_mnr, affix_rule arule, affix_rule args_rule,
				  affix_term_list aterms, guard_type type,
				  affix_term *tgt_var, member_list *guards)
{ /* Check if the actual argument is a single affix variable */
  affix_term first_term = aterms -> array[0];
  if (is_a_single_affix_variable (aterms) && equivalent_affix_rule (arule, first_term -> adef))
    { /* If the aterms are not a local variable, create a local to handle flow */
      /* Note: this code may not be used for transient affixes */
      if (is_a_global_variable (first_term) && (*tgt_var == affix_term_nil))
	*tgt_var = new_local_affix (first_term -> line, first_term -> col, arule);

      /* If we are forced to assign to a local, we do so for transients or globals */
      if (*tgt_var != affix_term_nil)
	{ affix_term src = (type == join)?first_term:*tgt_var;
	  affix_term dst = (type == join)?*tgt_var:first_term;
	  member new_assign = new_Assign (first_term -> line, first_term -> col,
					  attach_affix_term (src), attach_affix_term (dst));
	  *guards = new_member_list ();
	  app_member_list (*guards, new_assign);
	}
      else *tgt_var = attach_affix_term (first_term);
      return (1);
    };

  /* Check if we need an arithmetic split/join for this argument */
  if (is_a_primitive_type (args_rule) && equivalent_affix_rule (arule, args_rule))
    { /* Construct a local resolved guard */
      affix_term_list guard_parts = new_affix_term_list ();
      guard_type arith_type = (type == join)?arith_join:arith_split;
      int can_fail = (type == arith_split);
      member local_arith_guard;

      /* If we are forced to assign to a local, we do so for transients */
      /* If not, we need a local variable */
      if (*tgt_var == affix_term_nil)
	*tgt_var = new_local_affix (first_term -> line, first_term -> col, arule);

      /* Create local arithmetic guard */
      app_affix_term_list (guard_parts, attach_affix_term (first_term));
      local_arith_guard = new_Res_guard (first_term -> line, first_term -> col,
					 arith_type, attach_affix_term (*tgt_var),
					 0, can_fail, guard_parts);
      *guards = new_member_list ();
      app_member_list (*guards, local_arith_guard);
      return (1);
    };

  /* Last chance: we match by a split or join */
  if (is_a_primitive_type (arule)) return (0);
  if (match_terms (curr_mnr, aterms, arule, type, guards) == 1)
    { modify_argument_guards (type, tgt_var, guards);
      return (1);
    };

  /* No recognition, fail */
  return (0);
}

static int psig_matches_argument (int curr_mnr, psig ps, apar ap, call_arg carg)
{ affix_rule args_rule1, args_rule2;
  affix_term_list aterms1, aterms2;
  affix_rule arule = ps -> ftyp;
  affix_term aterm1, aterm2;

  /* Precomputation on apar */
  if (ap -> tag == TAGSingle)
    { aterms1 = ap -> Single.exp;
      aterm1 = aterms1 -> array[0];
      args_rule1 = ap -> Single.adef;
      aterms2 = aterms1;
      aterm2 = aterm1;
      args_rule2 = args_rule1;
    }
  else
    { aterms1 = ap -> Bidir.exp1;
      aterms2 = ap -> Bidir.exp2;
      aterm1 = aterms1 -> array[0];
      aterm2 = aterms2 -> array[0];
      args_rule1 = ap -> Bidir.adef1;
      args_rule2 = ap -> Bidir.adef2;
    };

  /* Switch on the direction of the affix flow */
  switch (ps -> fdir)
    { case d_inherited:
	{ if (ap -> tag != TAGSingle) return (0);
	  return (ftyp_matches_argument (curr_mnr, arule, args_rule1, aterms1, join,
					 &carg -> inh_term, &carg -> inh_guards));
	};
      case d_derived:
	{ if (ap -> tag != TAGSingle) return (0);
	  return (ftyp_matches_argument (curr_mnr, arule, args_rule2, aterms1, split,
					 &carg -> der_term, &carg -> der_guards));
	};
      case d_transient: 
	{ /* If the affix is a single local variable, we can use it as an actual transient */
	  /* In all other cases, especially when the actual is a global, we create a shadow */
	  if ((ap -> tag == TAGSingle) && is_a_single_affix_variable (aterms1) &&
	      !is_a_global_variable (aterm1)) ;
	  else
	    { /* Difficult case, create local variable to inherit and derive */
	      affix_term tgt_var = new_local_affix (aterm1 -> line, aterm1 -> col, arule);
	      carg -> inh_term = tgt_var;
	      carg -> der_term = attach_affix_term (tgt_var);
	    };
	  return (ftyp_matches_argument (curr_mnr, arule, args_rule1, aterms1, join,
					 &carg -> inh_term, &carg -> inh_guards) &&
		  ftyp_matches_argument (curr_mnr, arule, args_rule2, aterms2, split,
					 &carg -> der_term, &carg -> der_guards));
	};
      default: dcg_bad_tag (ps -> fdir, "psig_matches_argument");
    };
  return (0);
}

static int signature_matches_arguments (int curr_mnr, signature sig, apar_list args,
					call_arg_list cargs)
{ int ix;
  for (ix = 0; ix < sig -> size; ix++)
    { call_arg ca = new_call_arg ();
      psig ps = sig -> array[ix];
      apar ap = args -> array[ix];
      app_call_arg_list (cargs, ca);
      if (!psig_matches_argument (curr_mnr, ps, ap, ca))
        return (0);
    };
  return (1);
}

static void free_call_arg_parts (call_arg_list cargs)
{ int ix;
  for (ix = 0; ix < cargs -> size; ix++)
    { call_arg th = cargs -> array[ix];
      detach_affix_term (&th -> inh_term);
      detach_member_list (&th -> inh_guards);
      detach_affix_term (&th -> der_term);
      detach_member_list (&th -> der_guards);
    };
};

static int call_matches_decl (int curr_mnr, pdecl pd, apar_list p1args, apar_list p2args,
			      call_arg_list *sol_cargs1, call_arg_list *sol_cargs2)
{ signature p1sig = pd -> p1sig;
  signature p2sig = pd -> p2sig;
  call_arg_list cargs1 = init_call_arg_list (p1sig -> size);
  call_arg_list cargs2 = init_call_arg_list (p2sig -> size);
  if (signature_matches_arguments (curr_mnr, p1sig, p1args, cargs1) &&
      signature_matches_arguments (curr_mnr, p2sig, p2args, cargs2))
    { *sol_cargs1 = cargs1;
      *sol_cargs2 = cargs2;
      return (1);
    };
  free_call_arg_parts (cargs1);
  free_call_arg_parts (cargs2);
  detach_call_arg_list (&cargs1);
  detach_call_arg_list (&cargs2);
  return (0);
}

static void insert_call_guards (member_list mems, int *idx, dir d, call_arg_list cargs,
				variable_list pvars)
{ int ix;
  for (ix = 0; ix < cargs -> size; ix++)
    { call_arg carg = cargs -> array[ix];
 
      if (d == d_inherited)
	{ insert_guards (mems, idx, join, carg -> inh_guards);
	  add_created_locals_to_pvars (carg -> inh_guards, pvars);
	}
      else
	{ insert_guards (mems, idx, split, carg -> der_guards);
	  add_created_locals_to_pvars (carg -> der_guards, pvars);
	}
    };
}

static void add_call_vars (psig_list sig, call_arg_list cargs, affix_term_list args)
{ int ix;
  if (sig -> size != cargs -> size)
    dcg_internal_error ("add_call_vars");
  for (ix = 0; ix < sig -> size; ix++)
    { psig ps = sig -> array[ix];
      call_arg carg = cargs -> array[ix];
      if (ps -> fdir == d_inherited)
	app_affix_term_list (args, carg -> inh_term);
      else app_affix_term_list (args, carg -> der_term);
    };
}

static member create_resolved_call (int line, int col, pdecl pd, call_arg_list sol_cargs1,
				    call_arg_list sol_cargs2, module mod)
{ affix_term_list args1 = init_affix_term_list (sol_cargs1 -> size);
  affix_term_list args2 = init_affix_term_list (sol_cargs2 -> size);
  member new_call = new_Res_call (line, col, args1, args2);
  proc_type ptype = pd -> pdef -> pspec -> ptype;
  new_call -> Res_call.pdef = pd -> pdef;
  add_call_vars (pd -> p1sig, sol_cargs1, args1);
  add_call_vars (pd -> p2sig, sol_cargs2, args2);
  switch (ptype)
    { case prelude:
      case interlude:
      case postlude:
	contsens_error (mod, line, col, "cannot call %s %s", string_from_proc_type (ptype),
			pd -> pdef -> pspec -> canonic_name);
      default: break;
    };
  return (new_call);
}

static string make_error_pname (string_list pname_parts, int_list pname_chars, int p1size)
{ string ename = attach_string (pname_parts -> array[0]);
  int ix, cidx, fidx;
  int in_args = 0;

  for (ix = 1, cidx = 1, fidx = 0; ix < pname_chars -> size; ix++)
    if (pname_chars -> array[ix])
      { string ename_old = ename;
        ename = dcg_new_fmtd_string ("%s%s %s", ename_old, (in_args)?")":"",
				     pname_parts -> array[cidx]);
        detach_string (&ename_old);
        in_args = 0;
        cidx++;
      }
    else
      { int p2_switch = (fidx >= p1size);
	string ename_old = ename;
	if (!in_args)
          ename = dcg_new_fmtd_string ("%s (%s_", ename_old, (p2_switch)?"/":"");
        else ename = dcg_new_fmtd_string ("%s%s _", ename_old, (p2_switch)?"/":",");
	detach_string (&ename_old);
	in_args = 1;
	fidx++;
      };

  if (in_args)
    { string ename_old = ename;
      ename = dcg_new_fmtd_string ("%s)", ename_old);
      detach_string (&ename_old);
    };

  return (ename);
}

static void try_resolve_call (member_list mems, int *idx, pdecl_list pdecls,
			      variable_list pvars, module mod, string ename)
{ member m = mems -> array[*idx];
  apar_list p1args = m -> Call.p1args;
  apar_list p2args = m -> Call.p2args;
  call_arg_list sol_cargs1 = call_arg_list_nil;		/* Keep compiler happy */
  call_arg_list sol_cargs2 = call_arg_list_nil;
  int_list solutions = new_int_list ();
  pdecl sol_pd;
  int ix;

  /* Check every pdecl if it might match */
  for (ix = 0; ix < pdecls -> size; ix++)
    if (call_matches_decl (mod -> mnr, pdecls -> array[ix], p1args, p2args,
			   &sol_cargs1, &sol_cargs2))
      app_int_list (solutions, ix);
   
  /* Check for identification errors: either too many or no matching declarations */
  if (solutions -> size > 1)
    { contsens_error (mod, m -> line, m -> col, "call to '%s' is ambiguous", ename);
      dcg_wlog ("Candidates are: ");
      for (ix = 0; ix < solutions -> size; ix++)
	{ proc pdef = pdecls -> array[solutions -> array[ix]] -> pdef;
	  module dmod = all_modules -> array[pdef -> mnr];
	  spec dspec = pdef -> pspec;
	  dcg_wlog ("  module %s, line %d: %s %s", dmod -> mname, dspec -> line,
		    string_from_proc_type (dspec -> ptype), dspec -> canonic_name);
	};
      *idx += 1;
    }
  else if (!solutions -> size)
    { contsens_error (mod, m -> line, m -> col, "can't identify call to '%s'", ename);
      *idx += 1;
    }
  else
    { /*
         We have a solution: insert the guards for the inherited affixes
         before the call and those for the derived guards after it.
      */
      spec dspec;
      sol_pd = pdecls -> array[solutions -> array[0]];
      dspec = sol_pd -> pdef -> pspec;
      if (!equal_int_list (dspec -> pname_chars, m -> Call.pname_chars))
	contsens_warning (mod, m -> line, m -> col,
			  "call of %s has arguments and parts in different order",
			  dspec -> canonic_name);

      /* Resolve the remainder by inserting guards */
      insert_call_guards (mems, idx, d_inherited, sol_cargs1, pvars);
      insert_call_guards (mems, idx, d_inherited, sol_cargs2, pvars);
      mems -> array[*idx] = create_resolved_call (m -> line, m -> col, sol_pd,
						  sol_cargs1, sol_cargs2, mod);
      *idx += 1;
      insert_call_guards (mems, idx, d_derived, sol_cargs1, pvars);
      insert_call_guards (mems, idx, d_derived, sol_cargs2, pvars);
    };

  /* Cleanup */
  detach_int_list (&solutions);
}

static void check_call (member_list mems, int *idx, variable_list pvars, module mod)
{ member m = mems -> array[*idx];
  string_list pname_parts = m -> Call.pname_parts;
  apar_list p1args = m -> Call.p1args;
  apar_list p2args = m -> Call.p2args;
  string ename = make_error_pname (pname_parts, m -> Call.pname_chars, p1args -> size);
  pdecl_list pdecls;

  /* Do a precheck of the arguments as precomputation to resolution */
  check_apar_list (p1args, pvars, mod);
  check_apar_list (p2args, pvars, mod);

  /* Do a pre identification */
  pdecls = pre_identify_procedure (pname_parts, p1args -> size, p2args -> size, mod);
  if (pdecls == pdecl_list_nil)
    { contsens_error (mod, m -> line, m -> col, "can't identify call to '%s'", ename);
      *idx += 1;
    }
  else try_resolve_call (mems, idx, pdecls, pvars, mod, ename); 
  detach_string (&ename);
}

static void check_member (member_list mems, int *idx, variable_list pvars, module mod)
{ member m = mems -> array[*idx];
  switch (m -> tag)
    { case TAGCall:  check_call (mems, idx, pvars, mod); break;
      case TAGGuard: check_guard (mems, idx, pvars, mod); break;
      case TAGGroup: check_group (m -> Group.grp, pvars, mod);
		     /* Fall through */
      case TAGOp:    *idx += 1;
		     break;
      default: dcg_bad_tag (m -> tag, "check_member");
    };
}

void check_members (member_list mems, variable_list pvars, module mod)
{ int ix = 0;
  while (ix < mems -> size)
    check_member (mems, &ix, pvars, mod);
}

/*
   Check members2 is responsible a general simplification of all affix terms
   as well as the proper handling of implicit equals in split guard or calls.
*/
static void check_guard2 (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  simplify_affix_terms (m -> Res_guard.expr, p);
  *idx += 1;
}

static void check_member2 (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  switch (m -> tag)
    { case TAGRes_guard: check_guard2 (mems, idx, p); break;
      case TAGGroup:	 check_group2 (m -> Group.grp, p);
      case TAGRes_call:  /* fall thru */
      case TAGRes_equal:
      case TAGAssign:
      case TAGOp:
      case TAGCast:	 *idx += 1; break;
      default: dcg_bad_tag (m -> tag, "check_member2");
    };
}

void check_members2 (member_list mems, proc p)
{ int ix = 0;
  while (ix < mems -> size)
    check_member2 (mems, &ix, p);
}

/*
   Check members3 is responsible for reshaping guards for the following 3 cases:
   [ V -> V ]: Just remove and comment on silly programming
   [ M v1...v2 -> V ]: When vi === V, introduce temporary to replace V and assign after split.
   [ V -> M v1...vn ]: When vi === V, introduce temporary to replace vi and assign after split.
   [ V -> M vi...vn ]: When vi === vj (Implicit equal by consistent substitution), replace both
		       by temporaries, add an equality check + subsequent assignment
   Note: the second split restructuring must be tried before the first split restructuring.

   It is also responsible for restructuring calls that have common actual parameters at
   derived/transient positions. In this case the implicit equals must be resolved as well.
*/

/*
   Try restructure a guard when one of the component is also the target of the join.
   [ M v1...v2 -> V ] and at least one vi === V.

   Resolve by using a temporary to replace the target and assign the actual target
   after the join
*/
static int same_affix_term (affix_term var1, affix_term var2)
{ if (var1 -> tag != var2 -> tag) return (0);
  switch (var1 -> tag)
    { case TAGDyop:
	if (var1 -> Dyop.dop != var2 -> Dyop.dop) return (0);
	if (!same_affix_term (var1 -> Dyop.arg1, var2 -> Dyop.arg1)) return (0);
	return (same_affix_term (var1 -> Dyop.arg2, var2 -> Dyop.arg2));
      case TAGMonop:
	if (var1 -> Monop.mop != var2 -> Monop.mop) return (0);
	return (same_affix_term (var1 -> Monop.arg, var2-> Monop.arg));
      case TAGInt:
	return (var1 -> Int.num == var2 -> Int.num);
      case TAGSub:
	if (var1 -> Sub.vdef != var2 -> Sub.vdef) return (0);
	return (same_affix_term (var1 -> Sub.index, var2 -> Sub.index));
      case TAGVar:
        return (var1 -> Var.vdef == var2 -> Var.vdef);
      default: dcg_bad_tag (var1 -> tag, "same_affix_term");
    };
  return (0);
}

static int part_of_same_array (affix_term var1, affix_term var2)
{ if (var1 -> tag != TAGSub) return (0);
  if (var2 -> tag != TAGSub) return (0);
  return (var1 -> Sub.vdef == var2 -> Sub.vdef);
}

/*
   Try and replace [ M v1...vn -> T ] with vi === T for some i by
   [ M v1...vn -> Ttemp ], [ Ttemp -> T ]
*/
static void check_join3 (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  affix_term var = m -> Res_guard.var;
  affix_rule dst_arule = var -> adef;
  affix_term_list src_vars = m -> Res_guard.expr;
  affix_term new_local;
  member new_assign;
  int safe_guard = 1;
  int ix;

  /* Check if restructuring is necessary */
  for (ix = 0; ix < src_vars -> size; ix++)
    { affix_term src_var = src_vars -> array[ix];
      if (same_affix_term (var, src_var) || part_of_same_array (var, src_var))
	{ safe_guard = 0;
	  break;
	};
    };
  if (safe_guard)
    { *idx += 1;
      return;
    };

  /* Restructure the guard */
  new_local = new_local_affix (m -> line, m -> col, dst_arule);
  add_to_variable_list (p -> pvars, new_local -> Var.vdef);
  new_assign = new_Assign (m -> line, m -> col, attach_affix_term (new_local), var);
  m -> Res_guard.var = new_local;
  *idx += 1;
  ins_member_list (mems, *idx, new_assign);
  *idx += 1;
  contsens_hint_by_mnr (p -> mnr, m -> line, m -> col, "Restructured join guard (src/dst overlap)");
}

/*
   Since a second check is going to be made for splits, keep index unchanged
   Try and replace [ T -> M v1...vn ] where vi === vj for some i != j
   by [ T -> M v1..vi'..vj'..vn ], [ vi' = vj' ], [ vi' -> vi ] to solve implicit equals
*/
static void check_split3a (member_list mems, int idx, proc p)
{ member m = mems -> array[idx];
  affix_term_list dst_vars = m -> Res_guard.expr;
  int inconclusive = 0;
  int restructured = 0;
  int ix, iy;
  for (ix = 0; ix < dst_vars -> size; ix++)
    { affix_term dst_var1 = dst_vars -> array[ix];
      affix_term dst_var1_prime;
      int needs_restructuring = 0;
      int midx = idx + 1;
      member new_assign;
      for (iy = ix + 1; iy < dst_vars -> size; iy++)
	{ affix_term dst_var2 = dst_vars -> array[iy];
	  if (same_affix_term (dst_var1, dst_var2))
	    { needs_restructuring = 1;
	      break;
	    };
	  if (part_of_same_array (dst_var1, dst_var2))
	    inconclusive = 1;
	};
      if (!needs_restructuring) continue;

      /* We arrive here when we are certain that a restructuring is necessary */
      /* Create new variable vi' to replace vi and replace in original guard */
      dst_var1_prime = new_local_affix (m -> line, m -> col, dst_var1 -> adef);
      add_to_variable_list (p -> pvars, dst_var1_prime -> Var.vdef);
      dst_vars -> array[ix] = dst_var1_prime;
      
      /* Iterate over remaining members and replace identical terms with new locals */
      /* Insert explicit equal guards following the current guard */
      for (iy = ix + 1; iy < dst_vars -> size; iy++)
        { affix_term dst_var2 = dst_vars -> array[iy];
	  if (same_affix_term (dst_var1, dst_var2))
	    { affix_term dst_var2_prime = new_local_affix (m -> line, m -> col, dst_var2 -> adef);
	      member new_equal_guard;
	      add_to_variable_list (p -> pvars, dst_var2_prime -> Var.vdef);
	      dst_vars -> array[iy] = dst_var2_prime;
      	      new_equal_guard = new_Res_equal (m -> line, m -> col, 
					       attach_affix_term (dst_var1_prime),
					       attach_affix_term (dst_var2_prime));
	      ins_member_list (mems, midx, new_equal_guard);
	      midx++;
	    };
	};

      /* All equivalent destination vars have been replaced and implicit equals inserted */
      /* Now insert the final assignation to vi */
      new_assign = new_Assign (m -> line, m -> col, attach_affix_term (dst_var1_prime), dst_var1);
      ins_member_list (mems, midx, new_assign);
      restructured = 1;
    };

  if (inconclusive)
    contsens_warning_by_mnr (p -> mnr, m -> line, m -> col,
	"Unable to determine statically whether this split guard contains implicit equals.\n\
	 You should reformulate your code");
  if (restructured)
    contsens_warning_by_mnr (p -> mnr, m -> line, m -> col,
			     "Restructured split guard with implicit equal guards");
}

/*
   Try and replace [ T -> M v1...vn ] with vi === T for some i by
   [ T -> Ttemp ], [ Ttemp -> M v1...vn ]
*/
static void check_split3b (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  affix_term var = m -> Res_guard.var;
  affix_rule src_arule = var -> adef;
  affix_term_list dst_vars = m -> Res_guard.expr;
  affix_term new_local;
  member new_assign;
  int safe_guard = 1;
  int ix;
  for (ix = 0; ix < dst_vars -> size; ix++)
    { affix_term dst_var = dst_vars -> array[ix];
      if (same_affix_term (var, dst_var) || part_of_same_array (var, dst_var))
	{ safe_guard = 0;
	  break;
	};
    };
  if (safe_guard)
    { *idx += 1;
      return;
    };

  /* Restructure the guard */
  new_local = new_local_affix (m -> line, m -> col, src_arule);
  add_to_variable_list (p -> pvars, new_local -> Var.vdef);
  new_assign = new_Assign (m -> line, m -> col, var, attach_affix_term (new_local));
  m -> Res_guard.var = new_local;
  ins_member_list (mems, *idx, new_assign);
  *idx += 2;
  contsens_hint_by_mnr (p -> mnr, m -> line, m -> col,
			"Restructured split guard (src/dst) overlap)");
}

static void check_guard3 (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  switch (m -> Res_guard.typ)
    { case join:
	check_join3 (mems, idx, p);
	break;
      case split:
	check_split3a (mems, *idx, p);
	check_split3b (mems, idx, p);
	break;
      case arith_join:
      case arith_split: *idx += 1; break;
      default: dcg_bad_tag (m -> Res_guard.typ, "check_guard3");
    };
}

/*
   Check whether a call has derived/transient identical actual parameters
   If so implicit equal guards should be inserted after the call, followed
   by explicit assigns. We first implement detection.
*/
static void check_call3 (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  proc pdef = m -> Res_call.pdef;
  fpar_list p1fpars = pdef -> pspec -> p1fpars;
  affix_term_list p1args = m -> Res_call.p1args;
  int needs_restructuring = 0;
  int inconclusive = 0;
  int ix;
  for (ix = 0; ix < p1args -> size; ix++)
    { fpar fp1 = p1fpars -> array[ix];
      affix_term arg1 = p1args -> array[ix];
      int iy;

      if (fp1 -> fdir == d_inherited) continue;
      for (iy = ix + 1; iy > p1args -> size; iy++)
	{ fpar fp2 = p1fpars -> array[iy];
	  affix_term arg2 = p1args -> array[iy];
	  if (fp2 -> fdir == d_inherited) continue;

          if (same_affix_term (arg1, arg2)) needs_restructuring = 1;
	  else if (part_of_same_array (arg1, arg2)) inconclusive = 1;
	};
    };

  if (inconclusive)
    contsens_warning_by_mnr (p -> mnr, m -> line, m -> col,
	"Unable to determine statically whether this call contains implicit equals.\n\
	 You should reformulate your code");
  if (needs_restructuring)
    contsens_warning_by_mnr (p -> mnr, m -> line, m -> col,
	"This call contains implicit equal guards for some derived/transient parameters");
  *idx += 1;
}

static void check_assign3 (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  if (same_affix_term (m -> Assign.src, m -> Assign.dst))
    contsens_hint_by_mnr (p -> mnr, m -> line, m -> col, "Silly assignment");
  *idx += 1;
}

static void check_member3 (member_list mems, int *idx, proc p)
{ member m = mems -> array[*idx];
  switch (m -> tag)
    { case TAGRes_guard: check_guard3 (mems, idx, p); break;
      case TAGRes_call:  check_call3 (mems, idx, p); break;
      case TAGAssign:	 check_assign3 (mems, idx, p); break;
      case TAGGroup:	 check_group3 (m -> Group.grp, p); /* fall thru */
      case TAGRes_equal:
      case TAGOp:
      case TAGCast:	 *idx += 1; break;
      default: dcg_bad_tag (m -> tag, "check_member3");
    };
}

void check_members3 (member_list mems, proc p)
{ int ix = 0;
  while (ix < mems -> size)
    check_member3 (mems, &ix, p);
}
