/*
   File: options.c
   Defines flags and parameters of the mcdl compiler
   It also specifies some compiler defaults

   Copyright (C) 2008-2010 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: options.c,v 1.22 2012/01/20 09:41:49 marcs Exp $"
*/

/* global includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "options.h"

/* Exported flags and pragmats */
int interactive_compilation;
int partial_compilation;
int cdlc_compatible;
int strict_export;
int parse_only;

/* Exported code generation flags */
int generate_c_code;
int generate_object;
int optimize_code;
int no_code;

/* Exported debug flags */
int debug;
int dump_parser;
int dump_checker;
int dump_intermed;
int dump_optimizer;
int dump_flow;
int dump_affix_rules;

/* Local flags */
static int show_version;
static int show_include_path;
static int show_link_path;

/* exported module search path */
string_list module_search_path;

/* Source file names */
string dir_name;			/* Directory name */
string base_mname;			/* This should match the module name */
string source_name;			/* single file argument */
string output_fname;			/* implicit from opts */
string object_fname;			/* implicit from opts */
string c_code_fname;			/* implicit from opts */
string header_fname;			/* implicit from opts */

/* exported link configuration */
string_list library_search_path;
string_list link_library_names;

/* set default location of predef.cdl3 and libmrts */
#ifndef INCDIR
#define INCDIR "/home/marcs/share/mcdl/include"
#endif
#ifndef LIBDIR
#define LIBDIR "/home/marcs/lib"
#endif

/*
   Read a column separated path from an environment variable
   and store it into the paths list
*/
static void append_paths_from_environment (string_list paths, char *env_var)
{ char *env_paths = getenv (env_var);
  char buf[MAXPATHLEN+1];
  char *ptr, *dptr;

  /* Environment var has no value, ignore */
  if (env_paths == NULL) return;
  ptr = env_paths;
  if (!(*ptr)) return;
  dptr = buf;
	  
  /* As long as we have something to copy do so */
  while (1)
    if ((*ptr == ':') || !(*ptr))
      { *dptr = 0;
	if (strlen (buf))
	  { /* We have some path in the buffer */
#ifdef WIN32
	    struct _stat sbuf;
	    if ((_stat (buf, &sbuf) == 0) && (sbuf.st_mode & _S_IFDIR))
#else
	    struct stat sbuf;
	    if ((stat (buf, &sbuf) == 0) && S_ISDIR (sbuf.st_mode))
#endif
	    app_string_list (paths, new_string (buf));
	  };
	ptr++;
	if (!(*ptr)) return;
	dptr = buf;	
      }
    else *dptr++ = *ptr++;
}

void init_options ()
{ /* Compiler options */
  interactive_compilation = 0;
  partial_compilation = 0;
  cdlc_compatible = 0;
  strict_export = 0;
  parse_only = 0;

  /* Code generator options */
  generate_c_code = 0;
  generate_object = 0;
  optimize_code = 0;
  no_code = 0;

  /* Debug options */
  debug = 0;
  dump_parser = 0;
  dump_checker = 0;
  dump_intermed = 0;
  dump_optimizer = 0;
  show_version = 0;
  show_include_path = 0;
  show_link_path = 0;

  /* DCG options */
  full_verbose = 0;
  verbose = 0;

  /* Paths and names */
  dir_name = string_nil;
  base_mname = string_nil;
  source_name = string_nil;
  output_fname = string_nil;
  object_fname = string_nil;
  c_code_fname = string_nil;
  header_fname = string_nil;

  /*
     Initialize the search paths for modules and libraries
     and the list of all libraries to link against

     We will add the default paths after parsing the command line
  */
  module_search_path = new_string_list ();
  library_search_path = new_string_list ();
  link_library_names = new_string_list ();
}

static void print_usage ()
{ dcg_wlog ("usage: mcdl [flags] filename [more_flags]");
  dcg_wlog ("where flags and more_flags may be any of the following:");
  dcg_wlog ("   -h:  provide this help");
  dcg_wlog ("   -V:  show version");
  dcg_wlog ("   -d:  set debugging");
  dcg_wlog ("   -v:  verbose");
  dcg_wlog ("   -fv: full verbose");
  dcg_wlog ("   -po: only do context-free analysis of module");
  dcg_wlog ("   -dp: dump parser output");
  dcg_wlog ("   -dc: dump checker output");
  dcg_wlog ("   -df: dump flow analysis output");
  dcg_wlog ("   -di: dump intermediate code");
  dcg_wlog ("   -do: dump intermediate code after optimization");
  dcg_wlog ("   -da: dump affix_rules");
  dcg_wlog ("   -se: compile with strict exports");
  dcg_wlog ("   -ip: show default include path");
  dcg_wlog ("   -lp: show default link path");
  dcg_wlog ("   -cc: allow compatibility with old CDLC");
  dcg_wlog ("   -pc: compile as part, do not link");
  dcg_wlog ("   -c:  compile to object code, do not link");
  dcg_wlog ("   -C:  compile to C code");
  dcg_wlog ("   -O:  optimize generated code");
  dcg_wlog ("   -n:  do not generate code");
  dcg_wlog ("   -i:  enable interactive compilation");
  dcg_wlog ("   -o  oname: place output in file 'oname'");
  dcg_wlog ("   -ho oname: place header output in file 'oname'");
  dcg_wlog ("   -Ipath: add path to module search path");
  dcg_wlog ("   -Lpath: add path to library search path");
  dcg_wlog ("   -llib:  link object code also against this library");
  dcg_exit (4);
}

static void syntax_error (char *syn_error)
{ dcg_error (0, "error on command line: %s", syn_error);
  print_usage ();
}

static void scan_option (char *ptr, int *i, int argc, char **argv)
{ if (streq (ptr, "dp"))      { dump_parser = 1; verbose = 1; }
  else if (streq (ptr, "dc")) { dump_checker = 1; verbose = 1; }
  else if (streq (ptr, "df")) { dump_flow = 1; verbose = 1; }
  else if (streq (ptr, "di")) { dump_intermed = 1; verbose = 1; }
  else if (streq (ptr, "do")) { dump_optimizer = 1; verbose = 1; }
  else if (streq (ptr, "da")) { dump_affix_rules = 1; verbose = 1; }
  else if (streq (ptr, "d")) debug = 1;
  else if (streq (ptr, "V")) show_version = 1;
  else if (streq (ptr, "v"))  { show_version = 1; verbose = 1; }
  else if (streq (ptr, "fv")) { show_version = 1; verbose = 1; full_verbose = 1; }
  else if (streq (ptr, "po")) parse_only = 1;
  else if (streq (ptr, "se")) strict_export = 1;
  else if (streq (ptr, "ip")) show_include_path = 1;
  else if (streq (ptr, "lp")) show_link_path = 1;
  else if (streq (ptr, "cc")) cdlc_compatible = 1;
  else if (streq (ptr, "pc")) { partial_compilation = 1; generate_c_code = 1; }
  else if (streq (ptr, "C")) generate_c_code = 1;
  else if (streq (ptr, "c")) generate_object = 1;
  else if (streq (ptr, "O")) optimize_code = 1;
  else if (streq (ptr, "i")) interactive_compilation = 1;
  else if (streq (ptr, "h")) print_usage ();
  else if (streq (ptr, "n")) no_code = 1;
  else if (streq (ptr, "o"))
    { *i = *i + 1;
      if ((*i) < argc)
	{ detach_string (&output_fname);
	  output_fname = new_string (argv[*i]);
	}
      else syntax_error ("missing file name");
    }
  else if (streq (ptr, "ho"))
    { *i = *i + 1;
      if ((*i) < argc)
	{ detach_string (&header_fname);
	  header_fname = new_string (argv[*i]);
	}
      else syntax_error ("missing header file name");
    }
  else if (*ptr == 'I')
    { if (*(ptr + 1))
	app_string_list (module_search_path, new_string (ptr + 1));
      else
	{ *i = *i + 1;
	  if ((*i) < argc)
	    app_string_list (module_search_path, new_string (argv[*i]));
	  else syntax_error ("missing module search path component");
	};
    }
  else if (*ptr == 'L')
    { if (*(ptr + 1))
        app_string_list (library_search_path, new_string (ptr + 1));
      else
        { *i = *i + 1;
          if ((*i) < argc)
            app_string_list (library_search_path, new_string (argv[*i]));
          else syntax_error ("missing library search path component");
        };
    }
  else if (*ptr == 'l')
    { if (*(ptr + 1))
        app_string_list (link_library_names, new_string (ptr + 1));
      else syntax_error ("missing link library name");
    }
  else syntax_error ("illegal option specified");
}

#ifndef VERSION
#define VERSION "1.4"
#endif
static void report_version ()
{ if (!show_version) return;
  dcg_wlog ("This is MCDL, C Version %s, (C) M. Seutter", VERSION);
}

static void try_report_paths ()
{ if (show_include_path)
    { fprintf (stdout, "%s\n", INCDIR);
      dcg_exit (0);
    };
  if (show_link_path)
    { fprintf (stdout, "%s\n", LIBDIR);
      dcg_exit (0);
    };
}

/*
   Split the source name into a directory name and a basic (root) module name.
*/
static void set_basename_and_dirname ()
{ char buf[MAXSTRLEN];
  char *sptr, *dptr;
  char *last_slash;
  char *last_dot;

  /* Locate the last slash or DIR_SEP in the line */
  last_slash = NULL;
  for (sptr = source_name; (*sptr); sptr++)
    if ((*sptr == '/') || (*sptr == DIR_SEP)) last_slash = sptr;

  /* Determine the directory name and cut it from the source name */
  if (last_slash == NULL) dir_name = new_string (".");
  else if (last_slash == source_name)
    { dir_name = new_string ("/");
      source_name++;
    }
  else
    { size_t delta = last_slash - source_name;
      char buf2[MAXSTRLEN];
      strncpy (buf2, source_name, delta);
      buf2[delta] = '\0';
      dir_name = new_string (buf2);
      source_name = last_slash + 1;
    };

  /* Remember the last '.' in the line, while copying the source name */
  last_dot = NULL;
  for (sptr = source_name, dptr = buf; (*sptr); )
    { if (*sptr == '.') last_dot = dptr;
      *dptr++ = *sptr++;
    };

  /* Cut off the suffix, if it is there */
  if (last_dot != NULL) *last_dot = '\0';
  else *dptr = '\0';
  if (!buf[0])
    syntax_error ("empty module name");

  /* Set base name and add directory to search path */
  base_mname = new_string (buf);
}

/*
   After parsing the command line, we can add the defaults for the search paths
*/
static void add_default_search_paths ()
{ /*
     Add dir_name (the location where we search the main module)
     in the front of the search path, then append everything in the
     CDL_INCLUDE_PATH and finally append our default predef.cdl3
     search path at the end. If someone wants to overrule the
     predef.cdl3, that is fine with me: at your own risk... ;-)
  */
  ins_string_list (module_search_path, 0, attach_string (dir_name));
  append_paths_from_environment (module_search_path, "CDL_INCLUDE_PATH");
  app_string_list (module_search_path, new_string (INCDIR));

  /*
     Add everything in CDL_LIBRARY_PATH to the library search path
     and then append the default path for libmrts at the end
  */
  append_paths_from_environment (library_search_path, "CDL_LIBRARY_PATH");
  app_string_list (library_search_path, new_string (LIBDIR));

  /*
     Finally add -lmrts and -lm as libraries to link against
  */
  app_string_list (link_library_names, new_string ("mrts"));
  app_string_list (link_library_names, new_string ("m"));
}

void parse_command_line (int argc, char **argv)
{ int ix;
  for (ix = 1; ix < argc; ix++)
    { char *arg = argv[ix];
      if (arg[0] == '-') scan_option (arg+1, &ix, argc, argv);
      else if (source_name == string_nil) source_name = new_string (arg);
      else syntax_error ("too many arguments");
    };

  /* Try and report stuff */
  try_report_paths ();
  report_version ();

  /* Exit if no source given */
  if (source_name == string_nil)
    { if (!show_version)
	dcg_wlog ("no source name given, bailing out...");
      dcg_exit (0);
    };

  /* Set the basename for later use */
  set_basename_and_dirname ();
  add_default_search_paths ();
};
