/*
   File: parser.c
   Defines a parser for CDL3

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: parser.c,v 1.24 2011/08/13 15:57:30 marcs Exp $"
*/

/* global includes */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

/* libdcg includes */
#include <dcg.h>
#include <dcg_error.h>
#include <dcg_string.h>
#include <dcg_plist.h>
#include <dcg_plist_ops.h>

/* local includes */
#include "affix_rules.h"
#include "incr_error.h"
#include "options.h"
#include "lexer.h"
#include "cdl_ds.h"
#include "parser.h"

/* error administration */
static void parser_error (char *format, ...)
{ string fname;
  int line, col;
  char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  get_current_filename (&fname);
  get_current_position (&line, &col);
  dcg_error (0, "File %s, line %d, col %d: %s", fname, line, col, buf);
  maybe_enter_editor (fname, line, col);
};

static void parser_error_at (int line, int col, char *format, ...)
{ string fname;
  char buf[MAXSTRLEN];
  va_list arg_ptr;
  va_start (arg_ptr, format);
  vsprintf (buf, format, arg_ptr);
  va_end (arg_ptr);
  get_current_filename (&fname);
  dcg_error (0, "File %s, line %d, col %d: %s", fname, line, col, buf);
  maybe_enter_editor (fname, line, col);
};

/*
  Forward declarations to build parser CDL3 style
*/
static int is_module (string mname, module *mod);
static int is_root_module (string mname, module *mod);
static int is_plain_module (string mname, module *mod);
static int is_root_module_head (string *rname, variable_list *rvars);
static int is_plain_module_head (string mname, variable_list *rvars);
static void may_be_module_variables (variable_list *rvars);
static void should_be_module_element_sequence (variable_list vars);
static int is_module_element_sequence (variable_list vars);
static int is_module_element (variable_list vars);
static void may_be_module_interface (int export_ok, export_list *exports, string_list *ruses);
static void may_be_defines_part (int export_ok, export_list *exports);
static void should_be_specification_list (export_list *exports);
static int is_specification_list (export_list *exports);
static void should_be_specification (export *exp);
static int is_specification (export *exp);
static void may_be_uses_part (string_list *ruses);
static void should_be_module_name_list (string_list *ruses);
static int is_module_name_list (string_list *ruses);
static void should_be_module_body (proc_list procs, affix_rule_list arules);
static int is_module_body (proc_list procs, affix_rule_list arules);
static int is_definition (proc_list procs, affix_rule_list arules);
static int is_procedure (proc *rproc);
static int is_control_procedure (proc *rproc);
static int is_plain_procedure (proc *rproc);
static int is_control_procedure_head (spec *rspec);
static int is_control_type (proc_type *ptype);
static void should_be_procedure_name (string_list *pname_parts, int_list *pname_chars);
static int is_procedure_name (string_list pname_parts, int_list pname_chars);
static int is_plain_procedure_head (int is_def, spec *rspec);
static void may_be_real_procedure_type (int is_def, proc_type ptype, proc_type *real_ptype);
static void should_be_procedure_type (proc_type *ptype);
static int is_procedure_type (proc_type *ptype);
static void should_be_procedure_interface (string_list *pname_parts, int_list *pname_chars,
					   fpar_list *p1args, fpar_list *p2args);
static int is_procedure_interface (string_list pname_parts, int_list pname_chars,
				   fpar_list p1args, fpar_list p2args);
static void may_be_rest_procedure_interface (string_list pname_parts, int_list pname_chars,
					     fpar_list p1args, fpar_list p2args);
static int is_formal_parameters_pack (int *pass2_flag, int_list pname_chars,
				      fpar_list p1args, fpar_list p2args);
static void should_be_formal_parameters (int *pass2_flag, int_list pname_chars,
					 fpar_list p1args, fpar_list p2args);
static int is_formal_parameters (int *pass2_flag, int_list pname_chars,
				 fpar_list p1args, fpar_list p2args);
static int is_rest_formal_parameters (int *pass2_flag, int_list pname_chars,
				      fpar_list p1args, fpar_list p2args);
static void should_be_formal_parameter_list (int *pass2_flag, int_list pname_chars,
					     fpar_list p1args, fpar_list p2args);
static int is_formal_parameter_list (int *pass2_flag, int_list pname_chars,
				     fpar_list p1args, fpar_list p2args);
static void should_be_formal_parameter (fpar *rpar);
static int is_formal_parameter (fpar *rpar);
static void should_be_alternatives (alternative_list *alts);
static int is_alternatives (alternative_list *alts);
static void should_be_alternative (alternative *alt);
static int is_alternative (alternative *alt);
static void may_be_rest_sep_alternative (member_list p1mems, member_list p2mems);
static void should_be_rest_alternative (member_list p1mems, member_list p2mems);
static int is_rest_alternative (member_list p1mems, member_list p2mems);
static int is_group (member *mem);
static void may_be_rest_second_pass (member_list p2mems);
static void may_be_second_pass_group (member_list p2mems);
static int is_second_pass_group (member *mem);
static void should_be_second_pass_alternatives (alternative_list *alts);
static int is_second_pass_alternatives (alternative_list *alts);
static void should_be_second_pass_alternative (alternative *alt);
static int is_second_pass_alternative (alternative *alt);
static void may_be_rest_second_pass_alternative (member_list mems);
static void should_be_second_pass_member (member *mem);
static int is_second_pass_member (member *mem);
static void may_be_member_list (member_list mems);
static int is_member_list (member_list mems);
static void should_be_member (member *mem);
static int is_member (member *mem);
static int is_call (member *mem);
static void may_be_rest_call (string_list pname_parts, int_list pname_chars,
			      apar_list p1args, apar_list p2args);
static int is_actual_parameters_pack (int* pass2_flag, int_list pname_chars,
				      apar_list p1args, apar_list p2args);
static void should_be_actual_parameters (int *pass2_flag, int_list pname_chars,
					 apar_list p1args, apar_list p2args);
static int is_actual_parameters (int *pass2_flag, int_list pname_chars,
				 apar_list p1args, apar_list p2args);
static int is_rest_actual_parameters (int *pass2_flag, int_list pname_chars,
				      apar_list p1args, apar_list p2args);
static int is_actual_parameter_list (int *pass2_flag, int_list pname_chars,
				     apar_list p1args, apar_list p2args);
static void should_be_actual_parameter_list (int *pass2_flag, int_list pname_chars,
					     apar_list p1args, apar_list p2args);
static void should_be_actual_parameter (apar *rpar);
static int is_actual_parameter (apar *rpar);
static int is_guard (member *mem);
static int is_flow_operator (member *mem);
static void should_be_affix_expression (affix_term_list *terms);
static int is_affix_expression (affix_term_list *terms);
static int is_affix_term (affix_term *term);
static int is_affix_formula (affix_term *term);
static int is_affix_formula_of_prio (int prio, affix_term *term);
static void should_be_affix_primary (affix_term *term);
static int is_affix_primary (affix_term *term);
static int is_arithmetic_operator (int prio, operator *rop);
static int is_monadic_operator (operator *rop);
static void should_be_affix_variable (affix_term *term);
static int is_affix_variable (affix_term *term);
static int is_affix_rule (affix_rule_list arules);
static int rhs_is_single_affix_name (affix_alternative_list alts, string *syn_name);
static void expand_lhs_synonyms (affix_rule_list arules, int line, int col, string_list anames);
static int is_affix_rule_head (string_list *anames);
static void should_be_affix_alternatives (affix_alternative_list *alts);
static int is_affix_alternatives (affix_alternative_list *alts);
static void should_be_affix_alternative (affix_alternative *alt);
static int is_affix_alternative (affix_alternative *alt);
static int is_affix_element (affix_element *elem);

/*
  A.2 Programs and modules

  a) program:
       module sequence.
   
  b) module:
       root module;
       plain module.
   
  c) root module:
       root module head, uses part option, module body.
   
  d) plain module:
       plain module head, module interface, module body.
   
  e) root module head:
       "ROOT", module name, module variables option, ".".
   
  f) plain module head:
       "MODULE", module name, module variables option, ".".
*/
void should_be_module (string mname, module *mod)
{ if (is_module (mname, mod))
    { if (!end_of_file ())
	parser_error ("garbage at or after end of module");
      return;
    };
  parser_error ("module expected");
};

static int is_module (string mname, module *mod)
{ if (is_root_module (mname, mod)) return (1);
  return (is_plain_module (mname, mod));
}

static int is_root_module (string mname, module *mod)
{ export_list defines, exports;
  variable_list mvars;
  string_list uses;
  string fname;
  string rname;
  proc_list procs;
  affix_rule_list arules;
  if (!is_root_module_head (&rname, &mvars)) return (0);
  defines = new_export_list ();
  may_be_module_interface (partial_compilation, &exports, &uses);
  procs = new_proc_list ();
  arules = new_affix_rule_list ();
  should_be_module_body (procs, arules);
  get_current_filename (&fname);
  *mod = new_Root_module (dcg_attach (fname), dcg_attach (mname),
		          mvars, defines, uses, procs, arules, rname, exports);
  return (1);
}

static int is_plain_module (string mname, module *mod)
{ variable_list mvars;
  export_list defines;
  string_list uses;
  string fname;
  proc_list procs;
  affix_rule_list arules;
  if (!is_plain_module_head (mname, &mvars)) return (0);
  may_be_module_interface (1, &defines, &uses);
  procs = new_proc_list ();
  arules = new_affix_rule_list ();
  should_be_module_body (procs, arules);
  get_current_filename (&fname);
  *mod = new_Module (dcg_attach (fname), dcg_attach (mname),
	             mvars, defines, uses, procs, arules);
  return (1);
}

/*
   MS: module name is not the same as a procedure name
   Think about it
*/
static int is_root_module_head (string *rname, variable_list *rvars)
{ if (!is_keyword ("ROOT")) return (0);
  should_be_module_name (rname);
  may_be_module_variables (rvars);
  should_be_symbol (".");
  return (1);
}

static int is_plain_module_head (string mname, variable_list *rvars)
{ int line, col;
  string mname2;
  if (!is_keyword ("MODULE")) return (0);
  get_current_position (&line, &col);
  should_be_module_name (&mname2);
  if (!streq (mname, mname2))
    parser_error_at (line, col, "module name %s differs from file name %s", mname2, mname);
  may_be_module_variables (rvars);
  should_be_symbol (".");
  return (1);
}

/*  
  g) module variables:
	"=", module element sequence.

  h) module element:
	affix variable name, array declarator option;
        affix terminal.

  i) array declarator:
	"{", "}".
*/  
static void may_be_module_variables (variable_list *rvars)
{ *rvars = new_variable_list ();
  if (is_symbol ("="))
    should_be_module_element_sequence (*rvars);
}

static void should_be_module_element_sequence (variable_list vars)
{ if (is_module_element_sequence (vars)) return;
  parser_error ("module variable or affix terminal expected");
}

static int is_module_element_sequence (variable_list vars)
{ if (!is_module_element (vars)) return (0);
  while (is_module_element (vars)) ;
  return (1);
}

static int is_module_element (variable_list vars)
{ affix_term size;
  variable mvar;
  string aname;
  if (is_affix_terminal (&aname)) return (1);	/* Add free (aname)? */
  if (!is_affix_variable_name (&aname)) return (0);
  if (!is_symbol ("{"))
    { mvar = new_Module_var (aname);
      app_variable_list (vars, mvar);
      return (1);
    };
  if (is_affix_formula (&size))
    { should_be_symbol ("}");
      mvar = new_Module_array (aname, size);
    }
  else
    { should_be_symbol ("}");
      mvar = new_Module_array (aname, affix_term_nil);
    };
  app_variable_list (vars, mvar);
  return (1);
}

/* 
  A.2.1 Module interface
 
  a) module interface:
        defines part option, uses part option.
 
  b) defines part:
        "DEFINES", specification list, "."
 
  c) uses part:
        "USES", module name list, "."
 
  d) specification:
        procedure head;
        affix name.
 
  Note: the procedure head must be tried first.
*/ 
static void may_be_module_interface (int export_ok, export_list *exports, string_list *uses)
{ may_be_defines_part (export_ok, exports);
  may_be_uses_part (uses);
}

static void may_be_defines_part (int export_ok, export_list *exports)
{ if (is_keyword ("DEFINES"))
    { should_be_specification_list (exports);
      should_be_symbol (".");
      if (!export_ok)
	parser_error ("A defines list in a root module is only allowed for partial compilation");
    }
  else *exports = new_export_list ();
}

static void should_be_specification_list (export_list *exports)
{ if (is_specification_list (exports)) return;
  parser_error ("specification expected");
  *exports = export_list_nil;
}

static int is_specification_list (export_list *exports)
{ export exp;
  if (!is_specification (&exp)) return (0);
  *exports = new_export_list ();
  app_export_list (*exports, exp);
  while (is_symbol (","))
    { should_be_specification (&exp);
      app_export_list (*exports, exp);
    };
  return (1);
};

static void should_be_specification (export *rexp)
{ if (is_specification (rexp)) return;
  parser_error ("specification expected");
}

static int is_specification (export *rexp)
{ int line, col;
  string aname;
  spec rspec;
  get_current_position (&line, &col);
  if (is_plain_procedure_head (0, &rspec))
    { *rexp = new_Export_proc (line, col, rspec);
      return (1);
    };
  if (is_affix_name (&aname))
    { *rexp = new_Export_affix (line, col, aname);
      return (1);
    };
  return (0);

}

static void may_be_uses_part (string_list *ruses)
{ if (is_keyword ("USES"))
    { should_be_module_name_list (ruses);
      should_be_symbol (".");
    }
  else *ruses = new_string_list ();
}

static void should_be_module_name_list (string_list *ruses)
{ if (is_module_name_list (ruses)) return;
  parser_error ("module name expected");
  *ruses = string_list_nil;
}

static int is_module_name_list (string_list *ruses)
{ string mname;
  if (!is_module_name (&mname)) return (0);
  *ruses = new_string_list ();
  app_string_list (*ruses, mname);
  while (is_symbol (","))
    { should_be_module_name (&mname);
      add_uniquely_to_string_list (*ruses, mname);
    };
  return (1);
}

/* 
  A.2.3 Module body
 
  a) module body:
        definition sequence.
 
  b) definition:
        procedure;
        affix rule.

  Note: procedure should be tried first
        Some keywords may also be used as affix nonterminals
*/ 
static void should_be_module_body (proc_list procs, affix_rule_list arules)
{ if (is_module_body (procs, arules)) return;
  parser_error ("definition expected");
}

static int is_module_body (proc_list procs, affix_rule_list arules)
{ if (!is_definition (procs, arules)) return (0);
  while (is_definition (procs, arules)) ;
  return (1);
}

static int is_definition (proc_list procs, affix_rule_list arules)
{ proc aproc;
  if (is_procedure (&aproc))
    { app_proc_list (procs, aproc);
      return (1);
    };
  return (is_affix_rule (arules));
}

/* 
  A.3 Procedures
 
  a) procedure:
        control procedure;
        plain procedure.
 
  b) control procedure:
        control procedure head, ":", alternatives, ".".
 
  c) plain procedure:
        plain procedure head, ":", alternatives, ".";
	plain procedure head, ":", "EXTERNAL", text constant, ".".
*/ 
static int is_procedure (proc *rproc)
{ if (is_control_procedure (rproc)) return (1);
  return (is_plain_procedure (rproc));
}

static int is_control_procedure (proc *rproc)
{ alternative_list alts;
  spec cspec;
  if (!is_control_procedure_head (&cspec)) return (0);
  should_be_symbol (":");
  should_be_alternatives (&alts);
  should_be_symbol (".");
  *rproc = new_Alts (cspec, new_group (alts));
  return (1);
}

static int is_plain_procedure (proc *rproc)
{ alternative_list alts;
  string rname;
  spec pspec;
  if (!is_plain_procedure_head (1, &pspec)) return (0);
  should_be_symbol (":");
  if (is_keyword ("EXTERNAL"))
    { should_be_text_constant (&rname);
      should_be_symbol (".");
      *rproc = new_Ext_proc (pspec, rname);
      return (1);
    };
  should_be_alternatives (&alts);
  should_be_symbol (".");
  *rproc = new_Alts (pspec, new_group (alts));
  return (1);
}

/* 
  A.3.1 Procedure head
 
  a) control procedure head:
        control type, rule name.
 
  b) control type:
        "PRELUDE";
        "INTERLUDE";
        "POSTLUDE".
 
  c) rule name:
        name part, rule name,
        name part.
*/ 
static int is_control_procedure_head (spec *rspec)
{ string_list pname_parts;
  int_list pname_chars;
  proc_type ptype;
  fpar_list no_args;
  int line, col;
  get_current_position (&line, &col);
  if (!is_control_type (&ptype)) return (0);
  pname_parts = new_string_list ();
  pname_chars = new_int_list ();
  should_be_procedure_name (&pname_parts, &pname_chars);
  no_args = new_fpar_list ();
  *rspec = new_spec (line, col, ptype, pname_parts, pname_chars,
		     no_args, attach_fpar_list (no_args), string_nil, string_nil, ptype);
  return (1);
}

static int is_control_type (proc_type *rptype)
{ if (is_keyword ("PRELUDE"))        *rptype = prelude; 
  else if (is_keyword ("INTERLUDE")) *rptype = interlude;
  else if (is_keyword ("POSTLUDE"))  *rptype = postlude;
  else return (0);
  return (1);
}

static void should_be_procedure_name (string_list *pname_parts, int_list *pname_chars)
{ string error_name;
  *pname_parts = new_string_list ();
  *pname_chars = new_int_list ();
  if (is_procedure_name (*pname_parts, *pname_chars)) return;
  parser_error ("procedure name expected");
  get_erroneous_name (&error_name);
  app_string_list (*pname_parts, error_name);
  app_int_list (*pname_chars, 1);
}

static int is_procedure_name (string_list pname_parts, int_list pname_chars)
{ string part;
  if (!is_name_part (&part)) return (0);
  do
    { app_string_list (pname_parts, part);
      app_int_list (pname_chars, 1);
    }
  while (is_name_part (&part));
  return (1);
}

/*
  A.3.1 Rule head
 
  a) plain procedure head:
        procedure type, real procedure type option, procedure interface.
 
  We do not allow real procedure types in module interfaces.
*/ 
static int is_plain_procedure_head (int is_def, spec *rspec)
{ proc_type ptype, real_ptype;
  string_list pname_parts;
  int_list pname_chars;
  fpar_list p1args, p2args;
  int line, col;
  get_current_position (&line, &col);
  if (!is_procedure_type (&ptype)) return (0);
  may_be_real_procedure_type (is_def, ptype, &real_ptype);
  should_be_procedure_interface (&pname_parts, &pname_chars, &p1args, &p2args);
  *rspec = new_spec (line, col, ptype, pname_parts, pname_chars, p1args, p2args,
		     string_nil, string_nil, real_ptype);
  return (1);
}

/*  
  b) real procedure type:
        "[", procedure type, "]".
 
  c) procedure type:
        "ACTION";
	"PREDICATE";
        "PRED";
        "FUNCTION";  
        "TEST".
*/
static void may_be_real_procedure_type (int is_def, proc_type ptype, proc_type *real_ptype)
{ int line, col;
  get_current_position (&line, &col);
  if (is_symbol ("["))
    { should_be_procedure_type (real_ptype);
      should_be_symbol ("]");
      if (!is_def && !cdlc_compatible)
	parser_error_at (line, col, "real procedure types are not allowed in a DEFINES list");
    }
  else *real_ptype = ptype;
}

static void should_be_procedure_type (proc_type *rptype)
{ if (is_procedure_type (rptype)) return;
  parser_error ("procedure type expected");
  *rptype = none;
}

static int is_procedure_type (proc_type *rptype)
{ if (is_keyword ("ACTION"))         *rptype = action; 
  else if (is_keyword ("PREDICATE")) *rptype = predicate;
  else if (is_keyword ("PRED"))	     *rptype = predicate;
  else if (is_keyword ("FUNCTION"))  *rptype = function;
  else if (is_keyword ("TEST"))      *rptype = test;
  else return (0);
  return (1);
}

/*  
  d) procedure interface:
        name part, rest procedure interface option.
 
  e) rest procedure interface:
        name part, rest procedure interface option;
        formal parameters pack, rest procedure interface option.
*/ 
static void should_be_procedure_interface (string_list *pname_parts, int_list *pname_chars,
					   fpar_list *p1args, fpar_list *p2args)
{ string error_name;
  *pname_parts = new_string_list ();
  *pname_chars = new_int_list ();
  *p1args = new_fpar_list ();
  *p2args = new_fpar_list ();
  if (is_procedure_interface (*pname_parts, *pname_chars, *p1args, *p2args)) return;
  parser_error ("rule interface expected");
  get_erroneous_name (&error_name);
  app_string_list (*pname_parts, error_name);
  app_int_list (*pname_chars, 1);
}

static int is_procedure_interface (string_list pname_parts, int_list pname_chars,
				   fpar_list p1args, fpar_list p2args)
{ string pname_part;
  if (!is_name_part (&pname_part)) return (0);
  app_string_list (pname_parts, pname_part);
  app_int_list (pname_chars, 1);
  may_be_rest_procedure_interface (pname_parts, pname_chars, p1args, p2args);
  return (1);
}

static void may_be_rest_procedure_interface (string_list pname_parts, int_list pname_chars,
					     fpar_list p1args, fpar_list p2args)
{ int pass2_flag = 0;
  while (1)
    { string part;
      if (is_name_part (&part))
        { app_string_list (pname_parts, part);
	  app_int_list (pname_chars, 1);
    	}
      else if (is_formal_parameters_pack (&pass2_flag, pname_chars, p1args, p2args)) ;
      else if (cdlc_compatible &&
	       is_formal_parameter_list (&pass2_flag, pname_chars, p1args, p2args)) ;
      else break;
    };
}

/*
  f) formal parameters:
        "/", formal parameter list;
        rest formal parameters.
 
  g) rest formal parameters:
        formal parameter, ",", rest formal parameters;
        formal parameter, "/", formal parameter list.
        formal parameter.
*/
static int is_formal_parameters_pack (int *pass2_flag, int_list pname_chars,
				      fpar_list p1args, fpar_list p2args)
{ if (!is_symbol ("(")) return (0); 
  should_be_formal_parameters (pass2_flag, pname_chars, p1args, p2args);
  should_be_symbol (")");
  return (1);
}

static void should_be_formal_parameters (int *pass2_flag, int_list pname_chars,
					 fpar_list p1args, fpar_list p2args)
{ if (is_formal_parameters (pass2_flag, pname_chars, p1args, p2args)) return;
  parser_error ("formal parameters expected");
}

static int is_formal_parameters (int *pass2_flag, int_list pname_chars,
				 fpar_list p1args, fpar_list p2args)
{ if (is_symbol ("/"))
    { if (*pass2_flag)
        parser_error ("Multiple pass2 indicator '/'");
      *pass2_flag = 1;
      should_be_formal_parameter_list (pass2_flag, pname_chars, p1args, p2args);
      return (1);
    };
  return (is_rest_formal_parameters (pass2_flag, pname_chars, p1args, p2args));
}

static int is_rest_formal_parameters (int *pass2_flag, int_list pname_chars,
				      fpar_list p1args, fpar_list p2args)
{ fpar par;
  if (!is_formal_parameter (&par)) return (0);
  if (*pass2_flag) app_fpar_list (p2args, par);
  else app_fpar_list (p1args, par);
  app_int_list (pname_chars, 0);
  while (1)
    { if (is_symbol ("/"))
	{ if (*pass2_flag)
	    parser_error ("Multiple pass2 indicator '/'");
	  *pass2_flag = 1;
	  should_be_formal_parameter_list (pass2_flag, pname_chars, p1args, p2args);
	  return (1);
	};
      if (is_symbol (","))
	{ should_be_formal_parameter (&par);
	  if (*pass2_flag) app_fpar_list (p2args, par);
	  else app_fpar_list (p1args, par);
	  app_int_list (pname_chars, 0);
	}
      else break;
    };
  return (1);
}

/*
  h) formal parameter:
        input parameter;
        output parameter;
        transient parameter.
 
  i) input parameter: ">", affix variable name.
     output parameter: affix variable name, ">".
     transient parameter: ">", affix variable name, ">".
*/ 
static void should_be_formal_parameter_list (int *pass2_flag, int_list pname_chars,
					     fpar_list p1args, fpar_list p2args)
{ if (is_formal_parameter_list (pass2_flag, pname_chars, p1args, p2args)) return;
  parser_error ("formal parameter expected");
};

static int is_formal_parameter_list (int *pass2_flag, int_list pname_chars,
				     fpar_list p1args, fpar_list p2args)
{ fpar par;
  if (!is_formal_parameter (&par)) return (0);
  if (*pass2_flag) app_fpar_list (p2args, par);
  else app_fpar_list (p1args, par);
  app_int_list (pname_chars, 0);
  while (is_symbol (","))
    { should_be_formal_parameter (&par);
      if (pass2_flag) app_fpar_list (p2args, par);
      else app_fpar_list (p1args, par);
      app_int_list (pname_chars, 0);
    };
  return (1);
}

static void should_be_formal_parameter (fpar *rpar)
{ if (is_formal_parameter (rpar)) return;
  parser_error ("formal parameter expected");
  *rpar = fpar_nil;
}

static int is_formal_parameter (fpar *rpar)
{ int line, col;
  string fname;
  dir fdir;
  get_current_position (&line, &col);
  if (is_symbol (">"))
    { should_be_affix_variable_name (&fname);
      if (is_symbol (">")) fdir = d_transient;
      else fdir = d_inherited;
      *rpar = new_fpar (line, col, fname, fdir);
      return (1);
    };
  if (is_affix_variable_name (&fname))
    { should_be_symbol (">");
      *rpar = new_fpar (line, col, fname, d_derived);
      return (1);
    };
  return (0);
}

/* 
  A.3.2 Alternatives
 
  a) alternatives:
        alternative, ";", alternatives
        alternative.
 
  b) alternative:
        member list, second pass option;
        flow operator;
        second pass;
        member_list, ",", group;
        member_list, ",", flow operator.
 
  c) group:
        "(", alternatives, ")".
*/
static void should_be_alternatives (alternative_list *alts)
{ if (is_alternatives (alts)) return;
  parser_error ("alternative expected");
  *alts = alternative_list_nil;
}

static int is_alternatives (alternative_list *alts)
{ alternative alt;
  if (!is_alternative (&alt)) return (0);
  *alts = new_alternative_list ();
  app_alternative_list (*alts, alt);
  while (is_symbol (";"))
    { should_be_alternative (&alt);
      app_alternative_list (*alts, alt);
    };
  return (1);
}

static void should_be_alternative (alternative *alt)
{ if (is_alternative (alt)) return;
  parser_error ("alternative expected");
  *alt = alternative_nil;
}

static int is_alternative (alternative *alt)
{ member_list p1mems, p2mems;
  int line, col;
  member mem;
  get_current_position (&line, &col);
  if (is_member (&mem))
    { p1mems = new_member_list ();
      p2mems = new_member_list ();
      app_member_list (p1mems, mem);
      may_be_rest_sep_alternative (p1mems, p2mems);
    }
  else if (is_flow_operator (&mem))
    { p1mems = new_member_list ();
      p2mems = new_member_list ();
      app_member_list (p1mems, mem);
      may_be_rest_sep_alternative (p1mems, p2mems);
    }
  else if (is_symbol ("/"))
    { p1mems = new_member_list ();
      p2mems = new_member_list ();
      may_be_rest_second_pass (p2mems);
    }
  else return (0);

  /* Insert implicit success for pass1 if no pass1 members */
  if (p1mems -> size == 0) app_member_list (p1mems, new_Op (line, col, f_success));
  *alt = new_alternative (p1mems, p2mems); 
  return (1);
}

static void may_be_rest_sep_alternative (member_list p1mems, member_list p2mems)
{ if (is_symbol (","))
    should_be_rest_alternative (p1mems, p2mems);
  else if (is_symbol ("/"))
    may_be_rest_second_pass (p2mems);
}

static void should_be_rest_alternative (member_list p1mems, member_list p2mems)
{ if (is_rest_alternative (p1mems, p2mems)) return;
  parser_error ("member, flow operator or group expected");
}

static int is_rest_alternative (member_list p1mems, member_list p2mems)
{ member mem;
  if (is_member (&mem))
    { app_member_list (p1mems, mem);
      may_be_rest_sep_alternative (p1mems, p2mems);
    }
  else if (is_flow_operator (&mem))
    app_member_list (p1mems, mem);
  else if (is_group (&mem))
    app_member_list (p1mems, mem);
  else return (0);
  return (1);
}

static int is_group (member *mem)
{ int line, col;
  alternative_list alts;
  get_current_position (&line, &col);
  if (!is_symbol ("(")) return (0);
  should_be_alternatives (&alts);
  should_be_symbol (")");
  *mem = new_Group (line, col, new_group (alts));
  return (1);
}

/*
  A.3.3 Second pass

  a) second pass:
	"/", member list option, second pass group option.

  b) second pass group:
	"(", second pass alternatives, ")".

  c) second pass alternatives:
	second pass alternative;
	second pass alternative, ";", second pass alternatives.

  d) second pass alternative:
	member list, ",", second pass group;
	member list, ",", flow operator;
	flow operator.
*/
static void may_be_rest_second_pass (member_list p2mems)
{ int line, col;
  get_current_position (&line, &col);
  may_be_member_list (p2mems);
  may_be_second_pass_group (p2mems);

  /* Insert implicit pass2 succeed, if no pass2 member present */
  if (p2mems -> size == 0) app_member_list (p2mems, new_Op (line, col, f_success));
}

static void may_be_second_pass_group (member_list p2mems)
{ member mem;
  if (is_second_pass_group (&mem))
    app_member_list (p2mems, mem);
}

static int is_second_pass_group (member *mem)
{ alternative_list alts;
  int line, col;

  get_current_position (&line, &col);
  if (!is_symbol ("(")) return (0);
  should_be_second_pass_alternatives (&alts);
  should_be_symbol (")");
  *mem = new_Group (line, col, new_group (alts));
  return (1);
}

static void should_be_second_pass_alternatives (alternative_list *alts)
{ if (is_second_pass_alternatives (alts)) return;
  parser_error ("second pass alternative expected");
  *alts = alternative_list_nil;
}

static int is_second_pass_alternatives (alternative_list *alts)
{ alternative alt;
  if (!is_second_pass_alternative (&alt)) return (0);
  *alts = new_alternative_list ();
  app_alternative_list (*alts, alt);
  while (is_symbol (";"))
    { should_be_second_pass_alternative (&alt);
      app_alternative_list (*alts, alt);
    };
  return (1);
} 

static void should_be_second_pass_alternative (alternative *alt)
{ if (is_second_pass_alternative (alt)) return;
  parser_error ("second pass alternative expected");
  *alt = alternative_nil;
}

static int is_second_pass_alternative (alternative *alt)
{ member_list p2mems;
  int line, col;
  member mem;
  get_current_position (&line, &col);
  if (is_member (&mem))
    { p2mems = new_member_list ();
      app_member_list (p2mems, mem);
      may_be_rest_second_pass_alternative (p2mems);
    }
  else if (is_flow_operator (&mem))
    { p2mems = new_member_list ();
      app_member_list (p2mems, mem);
    }
  else return (0);
  *alt = new_alternative (new_member_list (), p2mems);
  return (1);
}

static void may_be_rest_second_pass_alternative (member_list mems)
{ member mem;
  while (is_symbol (","))
    { should_be_second_pass_member (&mem);
      app_member_list (mems, mem);
    };
}

static void should_be_second_pass_member (member *mem)
{ if (is_second_pass_member (mem)) return;
  parser_error ("second pass member, flow operator or group expected");
  *mem = member_nil;
}

static int is_second_pass_member (member *mem)
{ if (is_member (mem)) return (1);
  if (is_flow_operator (mem)) return (1);
  if (is_second_pass_group (mem)) return (1);
  return (0);
}

/* 
  A.3.4 Members
 
  a) member:
        call;
        guard.
 
  b) call:
        name part, rest call option
 
  c) rest call:
        name part, rest call option;
        actual parameters pack, rest call option.
 
  d) actual parameters:
        "/", actual parameter list;
        rest actual parameters.
 
  e) rest actual parameters:
        actual parameter, ",", rest actual parameters;
        actual parameter, "/", actual parameter list.
        actual parameter.
*/
static void may_be_member_list (member_list mems)
{ (void) is_member_list (mems);
}

static int is_member_list (member_list mems)
{ member mem;
  if (!is_member (&mem)) return (0);
  app_member_list (mems, mem);
  while (is_symbol (","))
    { should_be_member (&mem);
      app_member_list (mems, mem);
    };
  return (1);
}

static void should_be_member (member *mem)
{ if (is_member (mem)) return;
  parser_error ("member expected");
  *mem = member_nil;
}

static int is_member (member *mem)
{ if (is_call (mem)) return (1);
  return (is_guard (mem));
}

static int is_call (member *mem)
{ apar_list p1args, p2args;
  string_list pname_parts;
  string pname_part;
  int_list pname_chars;
  int line, col;
  get_current_position (&line, &col);
  if (!is_name_part (&pname_part)) return (0);

  /* We have a name part, setup name part admin */
  pname_parts = new_string_list ();
  pname_chars = new_int_list ();
  app_string_list (pname_parts, pname_part);
  app_int_list (pname_chars, 1);
  p1args = new_apar_list ();
  p2args = new_apar_list ();
  may_be_rest_call (pname_parts, pname_chars, p1args, p2args);
  *mem = new_Call (line, col, pname_parts, pname_chars, p1args, p2args);
  return (1);
}

static void may_be_rest_call (string_list pname_parts, int_list pname_chars,
			      apar_list p1args, apar_list p2args)
{ int pass2_flag = 0;
  while (1)
    { string part;
      if (is_name_part (&part))   
	{ app_string_list (pname_parts, part);
	  app_int_list (pname_chars, 1);
	}
      else if (is_actual_parameters_pack (&pass2_flag, pname_chars, p1args, p2args)) ;
      else break;
    }
}

static int is_actual_parameters_pack (int* pass2_flag, int_list pname_chars,
				      apar_list p1args, apar_list p2args)
{ if (!is_symbol ("(")) return (0);
  should_be_actual_parameters (pass2_flag, pname_chars, p1args, p2args);
  should_be_symbol (")");
  return (1);
}

static void should_be_actual_parameters (int *pass2_flag, int_list pname_chars,
					 apar_list p1args, apar_list p2args)
{ if (is_actual_parameters (pass2_flag, pname_chars, p1args, p2args)) return;
  parser_error ("actual parameters expected");
}

static int is_actual_parameters (int *pass2_flag, int_list pname_chars,
				 apar_list p1args, apar_list p2args)
{ if (is_symbol ("/"))
    { if (*pass2_flag)
        parser_error ("Multiple pass2 indicator '/'");
      *pass2_flag = 1;
      should_be_actual_parameter_list (pass2_flag, pname_chars, p1args, p2args);
      return (1);
    };
  return (is_rest_actual_parameters (pass2_flag, pname_chars, p1args, p2args));
}

static int is_rest_actual_parameters (int *pass2_flag, int_list pname_chars,
				      apar_list p1args, apar_list p2args)
{ apar par;
  if (!is_actual_parameter (&par)) return (0);
  if (*pass2_flag) app_apar_list (p2args, par);
  else app_apar_list (p1args, par);
  app_int_list (pname_chars, 0);
  while (1)
    { if (is_symbol ("/"))
	{ if (*pass2_flag)
	    parser_error ("Multiple pass2 indicator '/'");
	  *pass2_flag = 1;
	  should_be_actual_parameter_list (pass2_flag, pname_chars, p1args, p2args);
	  return (1);
	};
      if (is_symbol (","))
	{ should_be_actual_parameter (&par);
	  if (*pass2_flag) app_apar_list (p2args, par);
	  else app_apar_list (p1args, par);
	  app_int_list (pname_chars, 0);
	}
      else break;
    };
  return (1);
}

/*
  f) actual parameter:
        affix expression;
        affix expression, "|", affix expression.
*/
static void should_be_actual_parameter_list (int *pass2_flag, int_list pname_chars,
					     apar_list p1args, apar_list p2args)
{ if (is_actual_parameter_list (pass2_flag, pname_chars, p1args, p2args)) return;
  parser_error ("actual parameter expected");
};

static int is_actual_parameter_list (int *pass2_flag, int_list pname_chars,
				     apar_list p1args, apar_list p2args)
{ apar par;
  if (!is_actual_parameter (&par)) return (0);
  if (*pass2_flag) app_apar_list (p2args, par);
  else app_apar_list (p1args, par);
  app_int_list (pname_chars, 0);
  while (is_symbol (","))
    { should_be_actual_parameter (&par);
      if (pass2_flag) app_apar_list (p2args, par);
      else app_apar_list (p1args, par);
      app_int_list (pname_chars, 0);
    };
  return (1);
}

static void should_be_actual_parameter (apar *rpar)
{ if (is_actual_parameter (rpar)) return;
  parser_error ("actual parameter expected");
  *rpar = apar_nil;
}

static int is_actual_parameter (apar *rpar)
{ affix_term_list p1;
  if (!is_affix_expression (&p1)) return (0);
  if (is_symbol ("|"))
    { affix_term_list p2;
      should_be_affix_expression (&p2);
      *rpar = new_Bidir (p1, p2);
    }
  else *rpar = new_Single (p1);
  return (1);
}

/* 
  g) flow operator:
        failure operator;
        success operator;
        abort operator;
        exit operator.
 
  h) failure operator: "-".
     success operator: "+".
     abort operator:   "?".
     exit operator:    "!".
*/ 
static int is_flow_operator (member *mem)
{ int line, col;
  flow_op fop;
  get_current_position (&line, &col);
  if (is_symbol ("+")) fop = f_success;
  else if (is_symbol ("-")) fop = f_failure;
  else if (is_symbol ("?")) fop = f_abort;
  else if (is_symbol ("!")) fop = f_exit;
  else return (0);
  *mem = new_Op (line, col, fop);
  return (1);
}

/* 
  A.3.5 Guards
 
  a) guard:
        "[", confrontation, "]".
 
  b) confrontation:
        join; split; assign; equal.
 
  c) join:   affix expression, "->", affix variable.
     split:  affix variable, "->", affix expression.
     assign: affix variable, "->", affix variable.
     equal:  affix variable, "=", affix variable.
 
  Since it is absolutely impossible to distinguish the first 4 alternatives
  while parsing they are recognized in one syntactic parse and analyzed later.
*/ 
static int is_guard (member *mem)
{ guard_type typ = undetermined;
  affix_term_list lhs, rhs;
  int line, col;
  get_current_position (&line, &col);
  if (!is_symbol ("[")) return (0);
  should_be_affix_expression (&lhs);
  if (is_symbol ("=")) typ = equal;
  else if (!is_symbol ("->"))
    parser_error ("-> or = expected");
  should_be_affix_expression (&rhs);
  should_be_symbol ("]");
  *mem = new_Guard (line, col, typ, lhs, rhs);
  return (1);
}

/*
  A.3.6 Affix expressions
 
  a) affix expression:
        affix term sequence.
 
  b) affix term:
        affix terminal;
        affix formula.
 
  c) affix subexpression:
        affix subexpression, arithmetic operator, affix subexpression;
        affix part.

  Terminology should improve here. An affix term is the actual counterpart
  of the affix element in the formal affix rule definition.
*/
static void should_be_affix_expression (affix_term_list *terms)
{ if (is_affix_expression (terms)) return;
  parser_error ("affix expression expected");
  *terms = affix_term_list_nil;
}

static int is_affix_expression (affix_term_list *terms)
{ affix_term term;
  if (!is_affix_term (&term)) return (0);
  *terms = new_affix_term_list ();
  app_affix_term_list (*terms, term);
  while (is_affix_term (&term))
    app_affix_term_list (*terms, term);
  return (1);
}

static int is_affix_term (affix_term *term)
{ int line, col;
  string marker;
  get_current_position (&line, &col);
  if (is_affix_terminal (&marker))
    { *term = new_Terminal (line, col, marker);
      return (1);
    };
  return (is_affix_formula (term));
}

static void should_be_affix_formula (affix_term *term)
{ if (is_affix_formula (term)) return;
  parser_error ("affix formula expected");
  *term = affix_term_nil;
}

static int is_affix_formula (affix_term *term)
{ return (is_affix_formula_of_prio (1, term));
}

static void should_be_affix_formula_of_prio (int prio, affix_term *term)
{ if (is_affix_formula_of_prio (prio, term)) return;
  parser_error ("affix_formula of priority %d expected", prio);
  *term = affix_term_nil;
}

static int is_affix_formula_of_prio (int prio, affix_term *term)
{ int line, col;
  operator dop;
  if (prio >= 6) return (is_affix_primary (term));
  get_current_position (&line, &col);
  if (!is_affix_formula_of_prio (prio + 1, term)) return (0);
  while (is_arithmetic_operator (prio, &dop))
    { affix_term arg2;
      should_be_affix_formula_of_prio (prio + 1, &arg2);
      *term = new_Dyop (line, col, dop, *term, arg2);
    };
  return (1);
}

/*
  d) affix primary:
        affix variable;
        number;
        text constant;
        monadic operator, affix primary;
        "!", affix variable;
        "(", affix expression, ")".
*/
static void should_be_affix_primary (affix_term *term)
{ if (is_affix_primary (term)) return;
  parser_error ("affix primary expected");
  *term = affix_term_nil;
}

static int is_affix_primary (affix_term *term)
{ int inum, kind, line, col;
  string tconst;
  operator mop;
  real rnum;
  get_current_position (&line, &col);
  if (is_number (&kind, &inum, &rnum))
    { if (kind) *term = new_Real (line, col, rnum);
      else *term = new_Int (line, col, inum);
      return (1);
    };
  if (is_text_constant (&tconst))
    { *term = new_Text (line, col, tconst);
      return (1);
    };
  if (is_monadic_operator (&mop))
    { affix_term arg;
      should_be_affix_primary (&arg);
      *term = new_Monop (line, col, mop, arg);
      return (1);
    };
  if (is_symbol ("!"))
    { affix_term abs;
      should_be_affix_variable (&abs);
      *term = new_Concr (line, col, abs);
      return (1);
    };
  if (is_symbol ("("))
    { affix_term_list terms;
      should_be_affix_expression (&terms);
      should_be_symbol (")");
      *term = new_Enclosed (line, col, terms);
      return (1);
    };
  return (is_affix_variable (term));
}

/*
  e) arithmetic operator:
        "+"; "-"; "*"; "%"; "%%"; "<<"; ">>"; "&"; "||"; "^".

  e') monadic operator:
	"+"; "-"; "~".
*/
static int is_arithmetic_operator (int prio, operator *rop)
{ if (ahead_symbol ("->")) return (0);
  if (ahead_symbol ("=")) return (0);
  switch (prio)
    { case 1:
	if (is_symbol ("||")) { *rop = bitwise_or; return (1); }
	else if (is_symbol ("^")) { *rop = bitwise_xor; return (1); }
	break;
      case 2:
	if (is_symbol ("&")) { *rop = bitwise_and; return (1); }
	break;
      case 3:
	if (is_symbol ("<<")) { *rop = shift_left; return (1); }
	else if (is_symbol (">>")) { *rop = shift_right; return (1); }
	break;
      case 4:
	if (is_symbol ("+")) { *rop = plus; return (1); }
	else if (is_symbol ("-"))  { *rop = minus; return (1); }
	break;
      case 5:
	if (is_symbol ("*")) { *rop = times; return (1); }
	else if (is_symbol ("%%")) { *rop = modulo; return (1); }
	else if (is_symbol ("%")) { *rop = divides; return (1); }
      default: break;
    };
  return (0);
}

static int is_monadic_operator (operator *rop)
{ if (ahead_symbol ("->")) return (0);
  if (is_symbol ("+")) { *rop = plus; return (1); }
  if (is_symbol ("-")) { *rop = minus; return (1); }
  if (is_symbol ("~")) { *rop = bitwise_not; return (1); }
  return (0);
}

/*
  f) affix variable:
        affix variable name, index option.
 
  g) index:
        "{", affix formula, "}".
*/
static void should_be_affix_variable (affix_term *term)
{ if (is_affix_variable (term)) return;
  parser_error ("affix variable expected");
  *term = affix_term_nil;
}

static int is_affix_variable (affix_term *term)
{ string aname;
  int line, col;
  get_current_position (&line, &col);
  if (!is_affix_variable_name (&aname)) return (0);
  if (is_symbol ("{"))
    { affix_term index;
      should_be_affix_formula (&index);
      should_be_symbol ("}");
      *term = new_Sub (line, col, aname, index);
    }
  else *term = new_Var (line, col, aname);
  return (1);
}

/* 
  A.3.7 Affix rules
 
  a) affix rule:
	affix rule head, "::", EXTERNAL, ".".
        affix rule head, "::", affix alternatives, "."
        affix rule head, "=", affix alternatives, "."

  b) affix rule head:
        affix name list.
 
  c) affix alternatives:
        affix alternative, ";", affix alternatives;
        affix alternative.
 
  d) affix alternative:
        affix element sequence.
 
  e) affix element:
        affix variable name;
        affix terminal.
*/ 
static int is_affix_rule (affix_rule_list arules)
{ affix_alternative_list alts;
  string_list anames;
  affix_rule arule;
  string syn_name;
  int line, col;
  string aname;
  get_current_position (&line, &col);
  if (!is_affix_rule_head (&anames)) return (0);
  aname = attach_string (anames -> array[0]);
  if (is_symbol ("="))
    { should_be_affix_alternatives (&alts);
      should_be_symbol (".");
      if (rhs_is_single_affix_name (alts, &syn_name))
	{ detach_affix_alternative_list (&alts);
	  arule = new_Affix_synonym (line, col, aname, syn_name);
	}
      else arule = new_Affix_alts (line, col, aname, alts);
      app_affix_rule_list (arules, arule);
    }
  else if (is_symbol ("::"))
    { if (is_symbol ("EXTERNAL"))
        { should_be_symbol (".");
	  arule = new_Affix_prim (line, col, aname);
	}
      else
	{ /* We must introduce a concretized rhs with an anonymous name */
          string anonymous_name = new_anonymous_affix_name ();
	  affix_element_list elems = new_affix_element_list ();
	  affix_element elem = new_Affix_var (line, col, anonymous_name);
	  affix_alternative_list an_alts = new_affix_alternative_list ();
	  affix_alternative an_alt = new_affix_alternative (elems);
	  affix_rule anonymous_rule;
	  should_be_affix_alternatives (&alts);
          should_be_symbol (".");
	  if (rhs_is_single_affix_name (alts, &syn_name))
	    { detach_affix_alternative_list (&alts);
	      anonymous_rule = new_Affix_synonym (line, col, anonymous_name, syn_name);
	    }
	  else anonymous_rule = new_Affix_alts (line, col, anonymous_name, alts);
	  app_affix_rule_list (arules, anonymous_rule);
	  app_affix_element_list (elems, elem);
	  app_affix_alternative_list (an_alts, an_alt);
          arule = new_Affix_abstr (line, col, aname, an_alts);
        };
      app_affix_rule_list (arules, arule);
    }
  else parser_error ("= or :: expected");
  expand_lhs_synonyms (arules, line, col, anames);
  detach_string_list (&anames);
  return (1);
}

static int rhs_is_single_affix_name (affix_alternative_list alts, string *syn_name)
{ affix_element_list elems;
  affix_element elem;
  if (alts -> size != 1) return (0);
  elems = alts -> array[0] -> elems;
  if (elems -> size != 1) return (0);
  elem = elems -> array[0];
  if (elem -> tag != TAGAffix_var) return (0);
  *syn_name = attach_string (elem -> Affix_var.vname);
  return (1);
}

static void expand_lhs_synonyms (affix_rule_list arules, int line, int col, string_list anames)
{ int ix;
  for (ix = 1; ix < anames -> size; ix++)
    { string syn_name = attach_string (anames -> array[0]);
      string lhs_name = attach_string (anames -> array[ix]);
      affix_rule syn_rule = new_Affix_synonym (line, col, lhs_name, syn_name);
      app_affix_rule_list (arules, syn_rule);
    };
}

static int is_affix_rule_head (string_list *anames)
{ string aname;
  if (!is_affix_name (&aname)) return (0);
  *anames = new_string_list ();
  app_string_list (*anames, aname);
  while (is_symbol (","))
    { should_be_affix_name (&aname);
      app_string_list (*anames, aname);
    };
  return (1);
}

static void should_be_affix_alternatives (affix_alternative_list *alts)
{ if (is_affix_alternatives (alts)) return;
  parser_error ("affix alternative expected");
  *alts = affix_alternative_list_nil;
}

static int is_affix_alternatives (affix_alternative_list *alts)
{ affix_alternative alt;
  if (!is_affix_alternative (&alt)) return (0);
  *alts = new_affix_alternative_list ();
  app_affix_alternative_list (*alts, alt);
  while (is_symbol (";"))
    { should_be_affix_alternative (&alt);
      app_affix_alternative_list (*alts, alt);
    };
  return (1);
}

static void should_be_affix_alternative (affix_alternative *alt)
{ if (is_affix_alternative (alt)) return;
  parser_error ("affix alternative expected");
  *alt = affix_alternative_nil;
}

static int is_affix_alternative (affix_alternative *alt)
{ affix_element_list elems;
  affix_element elem;
  if (!is_affix_element (&elem)) return (0);
  elems = new_affix_element_list ();
  app_affix_element_list (elems, elem);
  while (is_affix_element (&elem))
    app_affix_element_list (elems, elem);
  *alt = new_affix_alternative (elems);
  return (1);
}

static int is_affix_element (affix_element *elem)
{ string affix;
  int line, col;
  get_current_position (&line, &col);
  if (is_affix_variable_name (&affix))
    { *elem = new_Affix_var (line, col, affix);
      return (1);
    }
  else if (is_affix_terminal (&affix))
    { *elem = new_Affix_term (line, col, affix);
      return (1);
    }
  return (0);
}
