/*
   File: procedures.c
   Checks the procedures

   Copyright (C) 2008 Marc Seutter

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   CVS ID: "$Id: procedures.c,v 1.31 2011/09/02 10:11:23 marcs Exp $"
*/

/* standard includes */
#include <stdio.h>
#include <string.h>

/* support lib includes */
#include <dcg.h>
#include <dcg_alloc.h>
#include <dcg_error.h>
#include <dcg_string.h>

/* local includes */
#include "options.h"
#include "cdl_ds.h"
#include "ast_utils.h"
#include "globals.h"
#include "contsens.h"
#include "modules.h"
#include "affix_rules.h"
#include "variables.h"
#include "members.h"
#include "procedures.h"

static string make_concatenated_name (string_list pname_parts, int_list pname_chars)
{ string cname = attach_string (pname_parts -> array[0]);
  int ix, cidx;

  for (ix = 1, cidx = 1; ix < pname_chars -> size; ix++)
    if (pname_chars -> array[ix])
      { string cname_old = cname;
        cname = dcg_new_fmtd_string ("%s_%s", cname_old, pname_parts -> array[cidx]);
        detach_string (&cname_old);
	cidx++;
      };
  return (cname);
}

static string make_canonical_name (string_list pname_parts, int_list pname_chars,
				   signature p1sig, signature p2sig)
{ string cname = attach_string (pname_parts -> array[0]);
  int ix, cidx, fidx;
  int in_args = 0;

  for (ix = 1, cidx = 1, fidx = 0; ix < pname_chars -> size; ix++)
    if (pname_chars -> array[ix])
      { string cname_old = cname;
        cname = dcg_new_fmtd_string ("%s%s %s", cname_old, (in_args)?")":"",
				 pname_parts -> array[cidx]);
	detach_string (&cname_old);
	in_args = 0;
        cidx++;
      }
    else
      { int p2_switch, pre_dir, post_dir;
        string cname_old = cname;
	affix_rule arule;
	string aname;
	psig ps;

	/* Check which formal parameter to add */
	if (fidx >= p1sig -> size) ps = p2sig -> array[fidx - p1sig -> size];
	else ps = p1sig -> array[fidx];
	p2_switch = (fidx == p1sig -> size);
	arule = ps -> ftyp;
	if (arule == affix_rule_nil) aname = "?ERROR?";
	else aname = ps -> ftyp -> aname;

	/* Check where to add directions */
	switch (ps -> fdir)
	  { case d_inherited: pre_dir = 1; post_dir = 0; break;
	    case d_derived:   pre_dir = 0; post_dir = 1; break;
	    case d_transient: pre_dir = 1; post_dir = 1; break;
	    default: pre_dir = 0; post_dir = 0;	/* error */
	  };

	if (!in_args)
	  cname = dcg_new_fmtd_string ("%s (%s%s%s%s", cname_old, (p2_switch)?"/":"",
				       (pre_dir)?">":"", aname, (post_dir)?">":"");
	else
	  cname = dcg_new_fmtd_string ("%s%s %s%s%s", cname_old, (p2_switch)?"/":",",
				       (pre_dir)?">":"", aname, (post_dir)?">":"");
	detach_string (&cname_old);
	in_args = 1;
	fidx++;
      };

  if (in_args)
    { string cname_old = cname;
      cname = dcg_new_fmtd_string ("%s)", cname_old);
      detach_string (&cname_old);
    };

  return (cname);
}

/*
   Phase1:
   Register all procedure definitions into a local procedure tree
   and identify all exported definitions: we do not allow reexports.
   Also generate the concatenated and canonical name for the procedures.

   Note1: This collection includes all control procedures as well.
          When such a procedure is identified in a call, an error results
	  Likewise, control procedures may never be exported.

   Note2: Procedures are identified by their typing alone, but during
   call resolution we need the direction of the affixes as well to
   resolve implicit guards used in actual arguments.
*/
static int equivalent_formal_typing (signature psig1, signature psig2)
{ int ix;
  if (psig1 -> size != psig2 -> size) return (0);
  for (ix = 0; ix < psig1 -> size; ix++)
    if (!equivalent_affix_rule (psig1 -> array[ix] -> ftyp, psig2 -> array[ix] -> ftyp))
      return (0);
  return (1);
}

/*
   Do a lookup of the list of procedure declarations with these name parts
   If it does not yet exist as tag in the tree, create it. If there is
   already a declaration with the same typing, we have a clash.
*/
static int register_procedure_with_sigs (proc pdef, signature p1sig, signature p2sig, module mod)
{ pdecl_list pdecls = enter_proc_tree (&mod -> pdefs, pdef -> pspec -> pname_parts);
  pdecl new_pd;
  int ix;
  for (ix = 0; ix < pdecls -> size; ix++)
    { pdecl pd = pdecls -> array[ix];
      if (equivalent_formal_typing (pd -> p1sig, p1sig) &&
	  equivalent_formal_typing (pd -> p2sig, p2sig))
	return (0);
    };

  /* Procedure is new with this name and signature.. */
  new_pd = new_pdecl ();
  new_pd -> p1sig = p1sig;
  new_pd -> p2sig = p2sig;
  new_pd -> pdef = pdef;
  app_pdecl_list (pdecls, new_pd);
  return (1);
}

static void register_procedure_for_module (proc pdef, module mod)
{ spec pspec = pdef -> pspec;
  signature p1sig, p2sig;
  variable_list pvars;

  /* Register procedure */
  pdef -> pnr = all_procedures -> size;
  pdef -> mnr = mod -> mnr;
  app_proc_list (all_procedures, pdef);
  switch (pspec -> ptype)
    { case prelude:   app_proc_list (mod -> preludes, pdef);   break;
      case interlude: app_proc_list (mod -> interludes, pdef); break;
      case postlude:  app_proc_list (mod -> postludes, pdef);  break;
      default: break;
    };

  /* Check formals and gather as variables */
  check_proc_formals (pspec, mod, &pvars, &p1sig, &p2sig);
  if ((p2sig -> size != 0) || (pspec -> ptype == interlude))
    pass2_error (mod, pdef -> pspec -> line, pdef -> pspec -> col);
  pdef -> pvars = pvars;

  /* Generate the concatenated and canonical name for this procedure */
  pspec -> concat_name = make_concatenated_name (pspec -> pname_parts, pspec -> pname_chars);
  pspec -> canonic_name = make_canonical_name (pspec -> pname_parts, pspec -> pname_chars,
					       p1sig, p2sig);
  
  /* Try to add to the symbol table */
  if (!register_procedure_with_sigs (pdef, p1sig, p2sig, mod))
    contsens_error (mod, pspec -> line, pspec -> col,
		    "Multiple conflicting definitions for procedure %s", pspec -> canonic_name);
}

static void register_procedures_for_module (module mod)
{ proc_list procs = mod -> procs;
  int ix;
  mod -> preludes = new_proc_list ();
  mod -> interludes = new_proc_list ();
  mod -> postludes = new_proc_list ();
  for (ix = 0; ix < procs -> size; ix++)
    register_procedure_for_module (procs -> array[ix], mod);
}

static proc try_identify_procedure (string_list pname_parts, signature p1sig, signature p2sig,
				    module mod)
{ pdecl_list pdecls = lookup_proc_tree (mod -> pdefs, pname_parts);
  int ix;
  if (pdecls == pdecl_list_nil) return (proc_nil);
  for (ix = 0; ix < pdecls -> size; ix++)
    { pdecl pd = pdecls -> array[ix];
      if (equivalent_formal_typing (pd -> p1sig, p1sig) &&
	  equivalent_formal_typing (pd -> p2sig, p2sig))
	return (pd -> pdef);
    };
  return (proc_nil);
}

static int affix_flow_matches (signature sig, fpar_list fpars)
{ int ix;
  if (sig -> size != fpars -> size) return (0);		/* Over specification */
  for (ix = 0; ix < sig -> size; ix++)
    if (sig -> array[ix] -> fdir != fpars -> array[ix] -> fdir)
      return (0);
  return (1);
}

/*
   Identify exported procedures
*/
static void export_procedures_for_module (module mod)
{ int real_export = partial_compilation && (mod -> tag == TAGRoot_module);
  export_list exports = (real_export)?mod -> Root_module.exports:mod -> defines;
  int ix;
  for (ix = 0; ix < exports -> size; ix++)
    { export exp = exports -> array[ix];
      signature p1sig, p2sig;
      spec pspec;
      proc pdef;
      if (exp -> tag != TAGExport_proc) continue;

      /* Obtain signatures from spec formals */
      pspec = exp -> Export_proc.pspec;
      check_spec_formals (pspec, mod, &p1sig, &p2sig);
      pspec -> concat_name = make_concatenated_name (pspec -> pname_parts, pspec -> pname_chars);
      pspec -> canonic_name = make_canonical_name (pspec -> pname_parts, pspec -> pname_chars,
						   p1sig, p2sig);
      pdef = try_identify_procedure (pspec -> pname_parts, p1sig, p2sig, mod);
      if (pdef == proc_nil)
        { contsens_error (mod, exp -> line, exp -> col,
			  "Could not identify exported procedure %s", pspec -> canonic_name);
	  continue;
	}

      /* Register the found signatures of the specification with the export */
      exp -> Export_proc.p1sig = p1sig;
      exp -> Export_proc.p2sig = p2sig;
      exp -> Export_proc.pdef = pdef;

      /* Check the affix flow of the procedure against spec */
      if (!affix_flow_matches (p1sig, pspec -> p1fpars) ||
	  !affix_flow_matches (p2sig, pspec -> p2fpars))
	{ contsens_error (mod, exp -> line, exp -> col,
			  "Exported procedure %s has different affix flow as implementation",
			  pspec -> canonic_name);
	  continue;
	};

      /* Check procedure type against spec */
      if (pdef -> pspec -> ptype != pspec -> ptype)
	{ contsens_error (mod, exp -> line, exp -> col,
			  "Exported procedure %s has a different type", pspec -> canonic_name);
	  continue;
        };

      if ((pspec -> ptype != pspec -> real_ptype) &&
	  (pdef -> pspec -> real_ptype != pspec -> real_ptype))
	contsens_error (mod, exp -> line, exp -> col,
			"Exported procedure %s has a different actual type", pspec -> canonic_name);

      if (!equal_int_list (pspec -> pname_chars, pdef -> pspec -> pname_chars))
	contsens_warning (mod, exp -> line, exp -> col,
	"Exported procedure %s has a different order of name parts than its implementation",
			  pspec -> canonic_name);

      /* If handling real export, mark procedure */
      if (real_export && (pdef -> tag == TAGAlts)) pdef -> Alts.exported = 1;
    };
}

static void register_and_export_procedures ()
{ int ix;
  dcg_hint ("      registering and exporting procedures");
  for (ix = 0; ix < all_modules -> size; ix++)
    { register_procedures_for_module (all_modules -> array[ix]);
      export_procedures_for_module (all_modules -> array[ix]);
    };
  dcg_hint ("      collected %d procedures", all_procedures -> size);
}

/*
   Phase2: import procedure definitions
	   check procedure bodies
*/
static void import_procedures_from_module (module imp, module mod)
{ export_list defines = imp -> defines;
  int ix;
  for (ix = 0; ix < defines -> size; ix++) 
    { export exp = defines -> array[ix];
      signature p1sig, p2sig;
      proc pdef;
      if (exp -> tag != TAGExport_proc) continue;
      pdef = exp -> Export_proc.pdef;
      if (pdef == proc_nil) continue;
      p1sig = exp -> Export_proc.p1sig;
      p2sig = exp -> Export_proc.p2sig;

      /* Try to add to the symbol table */
      if (!register_procedure_with_sigs (pdef, p1sig, p2sig, mod))
        contsens_error (mod, 0, 0, "import of procedure '%s' from module %s, line %d conflicts",
			pdef -> pspec -> canonic_name, imp -> mname, exp -> line);
    };
}

static void import_procedures_in_module (module mod)
{ string_list uses = mod -> uses;
  int ix;
  for (ix = 0; ix < uses -> size; ix++)
    { string uname = uses -> array[ix];
      module imp = identify_module (uname);
      import_procedures_from_module (imp, mod);
    };
}

/*
   Checking of members and alternatives

   Calls are preidentified by name and number of pass1 and pass2 parameters
*/
pdecl_list pre_identify_procedure (string_list pname_parts, int p1nr, int p2nr, module mod)
{ pdecl_list pdecls = lookup_proc_tree (mod -> pdefs, pname_parts);
  pdecl_list my_pdecls;
  int ix;
  if (pdecls == pdecl_list_nil) return (pdecls);
  my_pdecls = init_pdecl_list (pdecls -> size);
  for (ix = 0; ix < pdecls -> size; ix++)
    { pdecl pd = pdecls -> array[ix];
      if ((pd -> p1sig -> size == p1nr) && (pd -> p2sig -> size == p2nr))
	app_pdecl_list (my_pdecls, attach_pdecl (pd));
    };
  return (my_pdecls);
}

void check_group (group grp, variable_list pvars, module mod)
{ alternative_list alts = grp -> alts;
  int ix;
  for (ix = 0; ix < alts -> size; ix++)
    { alternative alt = alts -> array[ix];
      check_members (alt -> p1mems, pvars, mod);
      if (alt -> p2mems -> size != 0)
	pass2_error (mod, alt -> p2mems -> array[0] -> line, alt -> p2mems -> array[0] -> col);
      check_members (alt -> p2mems, pvars, mod);
    };
}

static void check_procedures_in_module (module mod)
{ proc_list procs = mod -> procs;
  int ix;
  for (ix = 0; ix < procs -> size; ix++)
    { proc p = procs -> array[ix];
      if (p -> tag != TAGAlts) continue;
      check_group (p -> Alts.grp, p -> pvars, mod);
    };
}

/*
   Create a parameterless call to a control procedure
*/
static void create_proc_call (proc pdef, member_list mems)
{ member call = new_Res_call (pdef -> pspec -> line, pdef -> pspec -> col,
			      new_affix_term_list (), new_affix_term_list ());
  call -> Res_call.pdef = pdef;
  app_member_list (mems, call);
}

static void create_proc_calls (proc_list pdefs, member_list mems)
{ int ix;
  for (ix = 0; ix < pdefs -> size; ix++)
    create_proc_call (pdefs -> array[ix], mems);
}

/*
   Create the root procedures for every prelude, postlude and root call
*/
static proc create_root_proc (member_list mems, char *rname)
{ /* Create correct name parts admin, there are no formal parameters */
  string pname_part = new_string (rname);
  string_list pname_parts = new_string_list ();
  int_list pname_chars = new_int_list ();
  spec root_spec;
  proc root_proc;

  app_string_list (pname_parts, pname_part);
  app_int_list (pname_chars, 1);
  root_spec = new_spec (0, 0, action, pname_parts, pname_chars, new_fpar_list (), new_fpar_list (),
			attach_string (pname_part), attach_string (pname_part), action);

  /* Create root procedure */
  root_proc = new_Root (root_spec, mems);
  root_proc -> mnr = root_module -> mnr;
  root_proc -> pnr = all_procedures -> size;
  root_proc -> pvars = new_variable_list ();
  app_proc_list (all_procedures, root_proc);
  return (root_proc);
}

static void create_prelude_proc ()
{ member_list mems = init_member_list (all_modules -> size * 2);
  int ix;
  for (ix = 0; ix < all_modules -> size; ix++)
    create_proc_calls (all_modules -> array[ix] -> preludes, mems);
  root_module -> Root_module.prelude = create_root_proc (mems, "PRELUDE");
}

/*
   Identify the ROOT procedure if possible.
   For partial compilation, we allow the absence of a ROOT algorithm
*/
static void identify_root_call ()
{ string_list sl = new_string_list ();
  string rname = root_module -> Root_module.rname;
  pdecl_list rdecls;
  proc rdef = proc_nil;
  app_string_list (sl, attach_string (rname));
  rdecls = pre_identify_procedure (sl, 0, 0, root_module);
  if ((rdecls != pdecl_list_nil) && rdecls -> size)
    { member_list mems = new_member_list ();
      spec rspec;
      rdef = rdecls -> array[0] -> pdef;
      rspec = rdef -> pspec;
      if (rspec -> ptype != action)
        contsens_error (root_module, rspec -> line, rspec -> col,
			"The ROOT algorithm %s is required to be an ACTION", rname); 
      create_proc_call (rdef, mems);
      root_module -> Root_module.root = create_root_proc (mems, "ROOT");
    }
  else if (!partial_compilation)
    { contsens_error (root_module, 0, 0, "Can't identify root call of %s", rname);
      return;
    };
  detach_string_list (&sl);
}

static void create_postlude_proc ()
{ member_list mems = init_member_list (all_modules -> size * 2);
  int ix;
  for (ix = 0; ix < all_modules -> size; ix++)
    create_proc_calls (all_modules -> array[ix] -> postludes, mems);
  root_module -> Root_module.postlude = create_root_proc (mems, "POSTLUDE");
}

static void import_and_check_procedures ()
{ int ix;
  dcg_hint ("      importing and checking procedures");
  for (ix = 0; ix < all_modules -> size; ix++)
    { import_procedures_in_module (all_modules -> array[ix]);
      check_procedures_in_module (all_modules -> array[ix]);
    };

  /* Create control procedures */
  create_prelude_proc ();
  identify_root_call ();
  create_postlude_proc ();
}

/*
   Second check does not flag errors, but simplifies expressions
   and guards for intermediate code generation.
*/
void check_group2 (group grp, proc p)
{ alternative_list alts = grp -> alts;
  int ix;
  for (ix = 0; ix < alts -> size; ix++)
    { alternative alt = alts -> array[ix];
      check_members2 (alt -> p1mems, p);
      check_members2 (alt -> p2mems, p);
    };
}

static void check_procedures2_in_module (module mod)
{ proc_list procs = mod -> procs;
  int ix;
  for (ix = 0; ix < procs -> size; ix++)
    { proc p = procs -> array[ix];
      if (p -> tag != TAGAlts) continue;
      check_group2 (p -> Alts.grp, p);
    };
}

static void check_procedures2 ()
{ int ix;
  dcg_hint ("      simplifying guards and text expressions");
  for (ix = 0; ix < all_modules -> size; ix++)
    check_procedures2_in_module (all_modules -> array[ix]);
}

/*
   Third check transforms guards and calls in dangerous situations
   (Implicit equals, auto assigns and source/destination overlaps (splits)).
*/
void check_group3 (group grp, proc p)
{ alternative_list alts = grp -> alts;
  int ix;
  for (ix = 0; ix < alts -> size; ix++)
    { alternative alt = alts -> array[ix];
      check_members3 (alt -> p1mems, p);
      check_members3 (alt -> p2mems, p);
    };
}

static void check_procedures3_in_module (module mod)
{ proc_list procs = mod -> procs;
  int ix;
  for (ix = 0; ix < procs -> size; ix++)
    { proc p = procs -> array[ix];
      if (p -> tag != TAGAlts) continue;
      check_group3 (p -> Alts.grp, p);
    };
}

static void check_procedures3 ()
{ int ix;
  dcg_hint ("      restructuring guards and calls");
  for (ix = 0; ix < all_modules -> size; ix++)
    check_procedures3_in_module (all_modules -> array[ix]);
}

void check_procedures ()
{ dcg_warning (0, "   checking procedures...");
  register_and_export_procedures ();
  dcg_panic_if_errors ();
  import_and_check_procedures ();
  dcg_panic_if_errors ();
  check_procedures2 ();
  check_procedures3 ();
}
